import os
import sys

from pytest_mproc import get_ip_addr

verbose = bool(os.environ.get("PTMPROC_VERBOSE", False) not in ['0', 'False', 'false', 'FALSE'])


def set_verbose(is_verbose: bool):
    """
    set verbosity of output

    :param is_verbose: if True, prints debug statements on actions taken
    """
    global verbose
    verbose = bool(is_verbose)


def debug_print(msg: str, *args):
    """
    Print only if verbose is True.  Is is recommended to use format strings just as in logging.* statements
    to avoid unnecessary pre-computation when a message is not printed (verbosity off)

    :param msg: string or format string
    :param args: args for format string if required
    """
    global verbose
    if verbose:
        msg = f"[{get_ip_addr()}] {msg}"
        print(msg % args)


def always_print(msg: str, *args, as_error: bool = False):
    """
    Always print a message (to stderr)

    :param msg: a string or format string
    :param args: args to go with format string if required
    :return:
    """
    msg = (msg % args) + '\n'
    if verbose:
        msg = f"[{get_ip_addr()}] {msg}"
    os.write(sys.stderr.fileno() if as_error else sys.stdout.fileno(), msg.encode('utf-8'))


import inspect
import functools
import logging
import os
import sys
import time
from multiprocessing.managers import BaseManager
from typing import Any, Dict, Tuple, Optional
from pytest_mproc.user_output import debug_print, always_print
from pytest_mproc import find_free_port, get_auth_key_hex, get_auth_key

# assert plugin  # import takes care of some things on import, but not used otherwise; here to make flake8 happy


logging.basicConfig(level=logging.DEBUG)


class FixtureManager(BaseManager):

    class NoneValue:
        pass

    NONE = NoneValue()

    class Value:

        def __init__(self, val: Any):
            self._val = val

        def value(self) -> Any:
            return self._val

    def __init__(self, addr: Tuple[str, int]):
        super().__init__(address=addr, authkey=get_auth_key())

    # noinspection PyAttributeOutsideInit
    def start(self, *args, **kwargs):
        debug_print(f"Starting {self.__class__.__qualname__} server {self.address}...")
        # server:
        self._fixtures: Dict[str, Any] = {}
        self.__class__.register("get_fixture", self._get_fixture)
        self.__class__.register("put_fixture", self._put_fixture)
        super().start(*args, **kwargs)
        debug_print(f"Started {self.__class__.__qualname__} server.")

    def connect(self):
        self.__class__.register("get_fixture")
        self.__class__.register("put_fixture")
        debug_print(f"Connecting {self.__class__.__qualname__} server {self.address}...")
        super().connect()
        debug_print(f"Connected {self.__class__.__qualname__} server.")

    def _put_fixture(self, name: str, value: Any) -> None:
        self._fixtures[name] = value

    def _get_fixture(self, name: str) -> Any:
        if name not in self._fixtures:
            return self.Value(FixtureManager.NONE)
        result = self._fixtures.get(name)
        return self.Value(result)


class Node:

    class Manager(FixtureManager):

        PORT = int(os.environ.get('PTMPROC_NODE_MGR_PORT', find_free_port()))
        _singleton = None

        def __init__(self, as_main: bool, port: int, name: str = "Node.Manager"):
            super().__init__(("127.0.0.1", port))
            if not as_main:
                debug_print(f"Connected [{name}]")

        @classmethod
        def singleton(cls) -> "Node.Manager":
            if cls._singleton is None:
                # noinspection PyBroadException
                try:
                    cls._singleton = cls(as_main=False, port=cls.PORT)
                    cls._singleton.connect()
                    cls._singleton._is_serving = False
                except (OSError, EOFError):
                    debug_print(f"Looks like no node manager already running, starting ...")
                    cls._singleton = cls(as_main=True, port=cls.PORT)
                    cls._singleton.start()
                    cls._singleton._is_serving = True
                except Exception as e:
                    raise SystemError(f"FAILED TO START NODE MANAGER") from e
            return cls._singleton

        @classmethod
        def shutdown(cls) -> None:
            # noinspection PyProtectedMember
            if cls._singleton is not None and cls._singleton._is_serving:
                cls._singleton.shutdown()
                cls._singleton = None


class Global:

    class Manager(FixtureManager):

        _singleton: Optional["Global.Maager"] = None

        def __init__(self, host: str, port: int):
            super().__init__((host, port))

        @classmethod
        def singleton(cls, address: Optional[Tuple[str, int]] = None, as_client: bool = False) -> "Global.Manager":
            if cls._singleton and cls._singleton.address[0] is not None:
                return cls._singleton
            elif address is None:
                raise SystemError("Attempt to get Global manager before start or connect")
            if as_client:
                host, port = address
                host = host.split('@', maxsplit=1)[-1]
                assert host is not None and port is not None, \
                    "Internal error: host and port not provided for global manager"
                try:
                    singleton = cls(host=host, port=port)
                    singleton.connect()
                    singleton._is_serving = False
                except Exception as e:
                    raise
            else:
                host, port = address
                host = host.split('@', maxsplit=1)[-1]
                singleton = cls(host=host, port=port)
                try:
                    singleton.start()
                    singleton._is_serving = True
                except:
                    raise
            cls._singleton = singleton
            return cls._singleton

        @classmethod
        def shutdown(cls) -> None:
            # noinspection PyProtectedMember
            if cls._singleton is not None and cls._singleton._is_serving:
                cls._singleton.shutdown()
                cls._singleton = None

    def __init__(self, config):
        super().__init__(config)


def global_fixture(**kwargs):
    import pytest
    if 'scope' in kwargs:
        raise pytest.UsageError("Cannot specify scope for 'glboal' fixtures; they are always mapped to 'session'")
    if "--cores" not in sys.argv:
        return pytest.fixture(scope='session', **kwargs)
    value = None

    def _decorator(func):

        # noinspection DuplicatedCode
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            nonlocal value
            assert Global.Manager._singleton is not None, "Global Manager did not start as expected"
            global_mgr = Global.Manager.singleton()
            # when serving potentially multiple full pytest session in a standalone server,
            # we distinguish sessions via the auth token
            # noinspection PyUnresolvedReferences
            key = f"{get_auth_key_hex()} {func.__name__}"
            value = value or global_mgr.get_fixture(key).value()
            if type(value) == FixtureManager.NoneValue:
                v = func(*args, **kwargs)
                if inspect.isgenerator(v):
                    try:
                        value = next(v)
                    except StopIteration:
                        raise Exception("Generator did not yield")
                else:
                    value = v
                # noinspection PyUnresolvedReferences
                global_mgr.put_fixture(key, value)
            return value

        return pytest.fixture(scope='session', **kwargs)(_wrapper)

    return _decorator


def node_fixture(**kwargs):
    import pytest
    if 'scope' in kwargs:
        raise pytest.UsageError("Cannot specify scope for 'glboal' fixtures; they are always mapped to 'session'")
    if "--cores" not in sys.argv:
        return pytest.fixture(scope='session', **kwargs)

    def _decorator(func):
        value = None

        # noinspection DuplicatedCode
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            nonlocal value
            node_mgr = Node.Manager.singleton()
            func_name = func.__name__
            # noinspection PyUnresolvedReferences
            value = value or node_mgr.get_fixture(func_name).value()
            if type(value) == FixtureManager.NoneValue:
                v = func(*args, **kwargs)
                if inspect.isgenerator(v):
                    try:
                        value = next(v)
                    except StopIteration:
                        raise Exception("Generator did not yield")
                else:
                    value = v
                # noinspection PyUnresolvedReferences
                node_mgr.put_fixture(func_name, value)
            return value

        return pytest.fixture(scope='session', **kwargs)(_wrapper)

    return _decorator



import asyncio
import hashlib
import multiprocessing
import os
import signal
import sys
import time
from contextlib import suppress
from enum import Enum
from glob import glob
from multiprocessing import Semaphore
from multiprocessing import JoinableQueue
from multiprocessing.managers import SyncManager
from pathlib import Path
from typing import Optional, Dict, List, Any, Tuple

import pytest

from pytest_mproc import get_auth_key, find_free_port, get_ip_addr
from pytest_mproc.coordinator import Coordinator
from pytest_mproc.fixtures import Global
from pytest_mproc.remote.ssh import SSHClient, remote_root_context
from pytest_mproc.user_output import always_print, debug_print


class Protocol(Enum):
    LOCAL = 'local'
    REMOTE = 'remote'
    SSH = 'ssh'
    DELEGATED = 'delegated'


class OrchestrationManager:

    class Value:
        def __init__(self, val):
            self._val = val

        def value(self):
            return self._val

    def __init__(self, host: str, port: int):
        Global.Manager.register("Queue", multiprocessing.Queue, exposed=["get", "put"])
        self._port = port
        self._host = host
        self._test_q: Dict[str, JoinableQueue] = {}
        self._results_q: Dict[str, JoinableQueue] = {}
        self._global_mgr = None
        # only for SSH protocol:
        self._clients: Dict[str, List[Coordinator]] = {}
        self._workers: Dict[str, Dict[str, Any]] = {}
        self._is_server = False
        self._mproc_pid: Optional[int] = None
        self._remote_pid: Optional[int] = None
        self._finalize_sem : Dict[str, multiprocessing.Semaphore] = {}

    @staticmethod
    def key() -> str:
        return hashlib.sha256(get_auth_key()).hexdigest()

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @staticmethod
    def create(uri: str, as_client: bool, project_name: Optional[str] = None) -> "OrchestrationManager":
        """
        create an orchestration manager to manage test & results q's as well as clients/workers
        one of three protocols can be used in uri: local for running a server local,
        remote for connecting to a remote server already running, os ssh (in which case host can be
        in form of a destination: user@host).  In the latter it will launch a remote instance
        to the specified host of he server.  Note that running locally means that a port
        must be opened for incoming requests.  Firewalls can prevent this, hence the other
        protocol options

        :param uri: URI in form <protocol>://host:port where protocol is one of 'local', 'remote', 'ssh'
        :param as_client: only for local protocol: whether to instantiate as a server or as a client
        :return: OrchestrationMenager instance
        """
        parts = uri.split('://', maxsplit=1)
        try:
            if uri.startswith('delegated://'):
                host = None
                port_text = uri.rsplit('://', maxsplit=1)[-1]
                if '@' in port_text:
                    username, port_text = port_text.split('@', maxsplit=1)
                protocol = Protocol.DELEGATED
            elif len(parts) == 1:
                host, port_text = parts[0].split(':', maxsplit=1)
                protocol = Protocol.LOCAL
            else:
                protocol_text = parts[0]
                host, port_text = parts[1].split(':', maxsplit=1)
                protocol = Protocol.LOCAL if as_client else Protocol(protocol_text)
            port = int(port_text) if port_text else (find_free_port() if protocol != Protocol.DELEGATED else -1)
        except ValueError:
            raise ValueError(f"URI {uri}: unknown protocol or invalid port specification")
        if protocol == Protocol.LOCAL:
            mgr = OrchestrationManager(host, port)
            if as_client:
                mgr.connect()
            else:
                mgr.start()
        elif protocol == Protocol.REMOTE:
            # always instantiate as client as there is an external server
            mgr = OrchestrationManager(host, port)
            mgr.connect()
        elif protocol == Protocol.SSH:
            assert project_name is not None
            mgr = OrchestrationManager(host, port)
            mgr._mproc_pid, mgr._remote_pid = mgr.launch(mgr.host, mgr.port, project_name)
            # wait for server to start listening
            time.sleep(3)
            mgr.connect()
        elif protocol == Protocol.DELEGATED:
            mgr = OrchestrationManager(host, port)
            # will connect later as server is delegated to one of the workers and must wait for that to come up
        else:
            raise SystemError("Invalid protocol to process")
        return mgr

    def Queue(self, size: Optional[int] = None):
        if size:
            return self._global_mgr.Queue(size)
        return self._global_mgr.Queue()

    def connect(self, host: Optional[str]=None) -> None:
        """
        Connect as a client to an existing server

        :param host: if host assignment defered, use this
        """
        if self._host is not None and host is not None:
            raise ValueError("Cannot specify host on connect when one already assigned on construction")
        elif self._host is None and host is None:
            raise ValueError(f"No host set")
        elif self._host is None:
            self._host = host
        Global.Manager.register("join")
        Global.Manager.register("get_finalize_sem")
        Global.Manager.register("register_client")
        Global.Manager.register("register_worker")
        Global.Manager.register("count")
        Global.Manager.register("completed")
        Global.Manager.register("JoinableQueue")
        Global.Manager.register("get_test_queue")
        Global.Manager.register("get_test_queue")
        Global.Manager.register("get_results_queue")
        host = self._host.split('@', maxsplit=1)[-1]
        debug_print(f"Connecting global mgr on {host} at {self._port} as client")
        self._global_mgr = Global.Manager.singleton(address=(host, self._port), as_client=True)

    def start(self):
        """
        Start this instance as the main server
        """
        assert self._global_mgr is None
        Global.Manager.register("join", self._join)
        Global.Manager.register("get_finalize_sem", self._get_finalize_sem)
        Global.Manager.register("register_client", self._register_client)
        Global.Manager.register("register_worker", self._register_worker)
        Global.Manager.register("count", self._count)
        Global.Manager.register("completed", self._completed)
        Global.Manager.register("JoinableQueue", JoinableQueue,
                                exposed=["put", "get", "task_done", "join", "close", "get_nowait"])
        Global.Manager.register("Semaphore", Semaphore, exposed=["acquire", "release"])
        Global.Manager.register("get_test_queue", self._get_test_queue)
        Global.Manager.register("get_results_queue", self._get_results_queue)
        # we only need these as dicts when running a standalone remote server handling multiple
        # pytest sessions/users
        self._global_mgr = Global.Manager.singleton(address=(self._host, self._port),
                                                    as_client=False)
        self._is_server = True

    def join(self, timeout: Optional[float] = None) -> None:
        with suppress(EOFError, ConnectionResetError, ConnectionRefusedError):
            self._global_mgr.join(self.key(), timeout)

    def _join(self, key: str,  timeout: Optional[float] = None) -> None:
        """
        For server, join all existing clients, presumably when expecting client to complete soon
        For client, do nothing

        :param timeout: Optional timeout to wait
        :raises: TimeoutError if timeout is specified and reached before join completes
        """
        for client in self._clients[key]:  # empty if client
            with suppress(ConnectionRefusedError):
                client.join(timeout=timeout)

    def register_client(self, client):
        """
        Register a client with this manager, to be joined and cleaned up at end of test run
        :param client: client to register
        """
        # noinspection PyUnresolvedReferences
        self._global_mgr.register_client(client, self.key())

    def register_worker(self, worker) -> None:
        """
        Register worker (there can be multiple workers per client)
        :param worker: worker to regisster
        """
        # noinspection PyUnresolvedReferences
        self._global_mgr.register_worker(worker, self.key())

    def get_finalize_sem(self):
        return self._global_mgr.get_finalize_sem(self.key())

    def count(self):
        # noinspection PyUnresolvedReferences
        return self._global_mgr.count(self.key()).value()

    def completed(self, host: str, index: int):
        # noinspection PyUnresolvedReferences
        return self._global_mgr.completed(host, index, self.key())

    def get_test_queue(self):
        # noinspection PyUnresolvedReferences
        return self._global_mgr.get_test_queue(self.key())

    def get_results_queue(self):
        # noinspection PyUnresolvedReferences
        return self._global_mgr.get_results_queue(self.key())

    def shutdown(self):
        if self._remote_pid is not None:
            with suppress(Exception):
                os.kill(self._remote_pid, signal.SIGINT)
            with suppress(Exception):
                os.kill(self._remote_pid, signal.SIGKILL)
        if self._mproc_pid is not None:
            with suppress(Exception):
                os.kill(self._mproc_pid, signal.SIGINT)
            with suppress(Exception):
                os.kill(self._mproc_pid, signal.SIGKILL)

    # server interface mapped to multiprocessing functions without the underscore in Global.Manager carried by this
    # instance
    def _get_test_queue(self, key: str):
        self._test_q.setdefault(key, JoinableQueue())
        return self._test_q[key]

    def _get_results_queue(self, key: str):
        self._results_q.setdefault(key, JoinableQueue())
        return self._results_q[key]

    # noinspection PyUnresolvedReferences
    def _register_client(self, client: "pytest_mproc.coordinator.Coordinator", key: str):
        self._clients.setdefault(key, []).append(client)

    def _register_worker(self, worker, key: str):
        ip_addr, pid = worker
        worker_key = f"{ip_addr}-{pid}"
        if key in self._workers and worker_key in self._workers[key]:
            raise Exception(f"Duplicate worker index: {worker} @  {worker_key}")
        self._workers.setdefault(key, {})[worker_key] = worker

    def _count(self, key: str):
        if key not in self._workers:
            return self.Value(0)
        return self.Value(len(self._workers[key]))

    def _completed(self, host: str, index: int, key: str):
        if key in self._workers:
            del self._workers[key][f"{host}-{index}"]
            if not self._workers[key]:
                del self._workers[key]

    def _get_finalize_sem(self, key: str):
        self._finalize_sem.setdefault(key, multiprocessing.Semaphore(0))
        return self._finalize_sem[key]

    @staticmethod
    def launch(host: str, port: int, project_name: str) -> Tuple[int, int]:
        destination = host
        host = destination.split('@')[-1]
        try:
            user, _ = destination.split('@', maxsplit=1)
        except ValueError:
            user = None
        sem = multiprocessing.Semaphore(0)
        mpmgr = SyncManager(authkey=get_auth_key())
        mpmgr.start()
        try:
            returns = mpmgr.dict()
            mproc = multiprocessing.Process(target=OrchestrationManager._main,
                                            args=(sem, host, port, user, returns, project_name))
            mproc.start()
            sem.acquire()
            mproc_pid = mproc.pid
            remote_proc_pid = returns.get('remote_proc_pid')
        finally:
            mpmgr.shutdown()
        del sem
        return mproc_pid, remote_proc_pid

    @staticmethod
    def _main(sem: multiprocessing.Semaphore, host: str, port: int, user: str, returns: Dict[str, int],
              project_name: str):
        asyncio.get_event_loop().run_until_complete(OrchestrationManager._main_server(
            sem, host, port, user, returns, project_name))

    @staticmethod
    async def _main_server(sem: multiprocessing.Semaphore, host: str, port: int,
                           user: str, returns: Dict[str, int], project_name: str):
        remote_proc = None
        try:
            ssh = SSHClient(host=host, username=user)
            root = Path(__file__).parent.parent
            files = glob(str(root / 'pytest_mproc' / '**' / '*.py'), recursive=True)
            async with remote_root_context(project_name=project_name, ssh_client=ssh, remote_root=None) as (tmpdir, venv_dir):
                remote_sys_executable = 'python{}.{}'.format(*sys.version_info)
                await ssh.execute_remote_cmd(remote_sys_executable, '-m', 'venv', 'venv',
                                             cwd=venv_dir)
                python = f"{str(venv_dir)}{os.sep}venv{os.sep}bin{os.sep}python3"
                # await ssh.execute_remote_cmd(python, '-m', 'pip', 'install', 'pytest')
                env = os.environ.copy()
                env['PYTHONPATH'] = f"{str(tmpdir)}:."
                env['AUTH_TOKEN_STDIN'] = '1'
                await ssh.mkdir(tmpdir / 'pytest_mproc')
                await ssh.mkdir(tmpdir / 'pytest_mproc' / 'remote')
                for f in files:
                    assert Path(f).exists()
                    await ssh.push(Path(f), tmpdir / Path(f).relative_to(root))
                debug_print(
                    f"Launching global manager: {python} {str(Path('pytest_mproc') / Path(__file__).name)} "
                    f"{host} {port}")
                remote_proc = await ssh.launch_remote_command(
                    python, str(tmpdir / Path('pytest_mproc') / Path(__file__).name),
                    host, str(port), project_name,
                    cwd=tmpdir,
                    env=env,
                    stderr=asyncio.subprocess.PIPE,
                    stdout=sys.stderr,
                    stdin=asyncio.subprocess.PIPE,
                    auth_key=get_auth_key()
                )
                import binascii
                remote_proc.stdin.write(binascii.b2a_hex(get_auth_key()) + b'\n')
                remote_proc.stdin.close()
                sem.release()
                returns.update({'remote_proc_id': remote_proc.pid})
                await remote_proc.wait()
        except:
            if remote_proc:
                remote_proc.terminate()
            sem.release()
            returns.update({'remote_proc_id': -1})
            raise


def main(host: str, port: int, project_name: str):
    if port == -1:
        port = find_free_port()
    sys.stdout.write(str(port) + '\n')
    sys.stdout.flush()
    always_print(f"Serving orchestration manager on {get_ip_addr()} {host}:{port} forever...")
    mgr = OrchestrationManager.create(uri=f"{host}:{port}", project_name=project_name,
                                      as_client=False)
    try:
        sem = mgr.get_finalize_sem()
        sem.acquire()
    finally:
        always_print("Shutting down delegated manager")
        mgr.shutdown()
    return mgr


if __name__ == "__main__":
    host = sys.argv[1]
    port = int(sys.argv[2])
    project_name = sys.argv[3]
    main(host, port, project_name)



import json
import logging
import multiprocessing
import os
import shutil
import socket
import sys
import time
from multiprocessing.process import current_process

import pytest
import subprocess
from asyncio import Semaphore
from pathlib import Path

from pytest_mproc import find_free_port
from pytest_mproc.main import RemoteExecutionThread
from pytest_mproc.orchestration import OrchestrationManager
from pytest_mproc.ptmproc_data import RemoteHostConfig, ProjectConfig

from pytest_mproc.remote.bundle import Bundle
from pytest_mproc.remote.ssh import SSHClient, CommandExecutionFailure

_base_dir = Path(__file__).parent.parent
logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__file__)
ssh_pswdless = (Path.home() / ".ssh" / "authorized_keys").exists()


class TestSSHClient:

    @pytest.mark.asyncio
    async def test_mkdtmp(self):
        ssh_client = SSHClient(socket.gethostbyname(socket.gethostname()))
        async with ssh_client.mkdtemp() as tmpdir:
            assert Path(tmpdir).exists()  # this is localhost after all
        assert not Path(tmpdir).exists()


@pytest.fixture(scope='session')
def bundle(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("remote_testing")
    with Bundle.create(
            root_dir=Path(str(tmpdir)),
            project_config=ProjectConfig(
                project_root=Path(__file__).parent.parent,
                src_paths=[Path("src"), Path("testsrc")],
                test_files=[Path("test/*.py")],
                project_name="pytest_mproc_test"
            ),
            system_executable=sys.executable
    ) as bundle:
        bundle.add_file(_base_dir / "test" / "run_test.sh", "scripts/run_test.sh")
        yield bundle


@pytest.mark.skipif(not ssh_pswdless, reason="No permission to ssh to localhost without password input")
@pytest.mark.asyncio
async def test_execute_bundle(bundle):
    lines = []
    try:
        async for line in bundle.monitor_remote_execution("localhost", ".", "-k", "alg1", timeout=100,
                                                          env=dict(os.environ)):
            lines.append(line)
            print(line)
    except CommandExecutionFailure:
        text = '\n'.join(lines)
        assert False, f"Execution failed:\n{text}"


@pytest.mark.asyncio
async def test_execute_remote_multi(bundle):
    hosts = [
        RemoteHostConfig("localhost", {"cores": "1"}),
        RemoteHostConfig("localhost"),
        RemoteHostConfig("localhost", {"cores": "1"}),
    ]
    sys.argv.extend(['--cores', '2'])
    sem = Semaphore(0)
    environ = {'PTMPROC_NODE_MGR_PORT': str(find_free_port()), 'PTMPROC_GLOBAL_MGR_PORT': str(find_free_port())}
    procs = await bundle.execute_remote_multi(hosts, sem, ".", "-k", "alg3", "--cores", "2",
                                              deploy_timeout=100, timeout=200,
                                              env=environ, auth_key=multiprocessing.current_process().authkey)
    for host, proc in procs.items():
        assert host == "localhost"
        assert proc.returncode != 0  # alg3 delibrately coded to fail


def test_remote_execution_cli(tmp_path):
    root = Path(__file__).parent.parent
    project_config = ProjectConfig(src_paths=[Path("src"), Path("testsrc")],
                                   test_files=[Path("test") / "*.py", Path("requirements.txt")],
                                   project_root=root,
                                   project_name="pytest_mproc_test"
                                   )
    project_config_path = tmp_path / "project.cfg"
    for path in project_config.src_paths + [Path("test")]:
        shutil.copytree(root / path, tmp_path / path.name)
    shutil.copy(root / "test" / "resources" / "requirements.txt", tmp_path / "requirements.txt")
    with open(project_config_path, 'w') as out:
        converted = {'src_paths': [str(p) for p in project_config.src_paths],
                     'test_files': [str(p) for p in project_config.test_files],
                     'project_name': project_config.project_name}
        out.write(json.dumps(converted))
        out.flush()
        # remote_host = 'fssh://pi@10.220.45.119:{find_free_port()}'
        remote_host = f'127.0.0.1:{find_free_port()}'
        client_connect = remote_host.split(":")[0]
        remote_server = f'delegated://pi@'
        client_connect = '10.220.45.110'
        # remote_server = remote_host
        args = [
            'pytest', '-s', 'test/test_mproc_runs.py', '-k', 'alg2',
            '--cores', '3',
            '--as-main', f"{remote_server}",
            '--project-structure', str(project_config_path),
            '--remote-worker', client_connect,
            '--remote-worker', client_connect,
            '--remote-worker', client_connect,
            '--remote-worker', client_connect,
            '--remote-worker', client_connect,
        ]
        sys.path.insert(0, str((Path(__file__).parent / "src").absolute()))
        env = os.environ.copy()
        env['PYTHONPATH'] = "./src"
        print(f">>>>>>>>  RUNNING {' '.join(args)}")
        completed = subprocess.run(args, stdout=sys.stdout, stderr=sys.stderr, timeout=1200, env=env,
                                   cwd=str(tmp_path))
        assert completed.returncode == 0, f"FAILED TO EXECUTE pytest from \"{' '.join(args)}\" from " \
                                          f"{str(Path(__file__).parent.absolute())}"


def test_remote_execution_thread(tmp_path, chdir):
    chdir.chdir(Path(__file__).parent.parent)
    root = Path(__file__).parent.parent
    project_config = ProjectConfig(
        src_paths=[Path("src"), Path("testsrc"), Path("resources") / "requirements.txt"],
        test_files=[Path("test/*.py")],
        project_root=root,
        project_name="pytest_mproc_test"
    )
    args = list(sys.argv)
    sys.argv = ["test/test_mproc_runs.py", "-k", "alg2", "--cores", "4"]
    ipname = socket.gethostbyname(socket.gethostname())
    port = find_free_port()
    command = [
        shutil.which("python3"), "-m", "pytest", "--as-main", f"{ipname}:{port}",
    ]
    command.extend(sys.argv)
    env = os.environ.copy()
    import binascii
    env['AUTH_TOKEN_STDIN'] = '1'
    env['PYTHONPATH'] = "src"
    main_proc = subprocess.Popen(command, stdout=sys.stdout, stderr=sys.stderr, cwd=str(Path(__file__).parent.parent),
                                 env=env, stdin=subprocess.PIPE)
    main_proc.stdin.write(binascii.b2a_hex(current_process().authkey) + b'\n')
    main_proc.stdin.close()
    client_hosts = [
        RemoteHostConfig(ipname),
        RemoteHostConfig(ipname),
        RemoteHostConfig(ipname),
    ]
    thread = RemoteExecutionThread(project_config=project_config,
                                   remote_sys_executable=shutil.which("python3"),
                                   remote_hosts_config=client_hosts,
                                   )
    try:
        sem = Semaphore(0)
        time.sleep(3)
        mgr = OrchestrationManager(host=ipname, port=port)
        # noinspection PyUnresolvedReferences
        results_q = mgr.get_results_queue()
        thread.start_workers(
            server=ipname,
            server_port=port,
            auth_key=current_process().authkey,
            finish_sem=sem,
            result_q=results_q,
        )
        sem.release()
        sem.release()
        sem.release()
        thread.join(timeout=3*240)
    finally:
        sys.argv = args
        try:
            assert main_proc.wait(timeout=5) == 0
        except (subprocess.TimeoutExpired, TimeoutError):
            main_proc.terminate()
            raise Exception("Main process failed to terminate")



"""
SSH utilities for remote client interactions
"""
import asyncio
import logging
import os
import shlex
import sys
import time
from contextlib import suppress
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import (
    AsyncIterator,
    Callable,
    Dict,
    Optional,
    TextIO,
    Tuple,
    Union, AsyncContextManager, List,
)

from pytest_mproc import user_output
from pytest_mproc.user_output import debug_print

SUCCESS = 0


class CommandExecutionFailure(Exception):
    """
    Raised on failure to execute SSH command on remote client
    """

    def __init__(self, cmd: str, rc: int):
        super().__init__(f"Execution of '{cmd}' return error status {rc}")
        self._rc = rc

    @property
    def rc(self):
        return self._rc


@dataclass
class SSHClient:
    """
    SSH Client objct for interacting with a specific client
    """
    host: str
    username: Optional[str] = None
    password: Optional[str] = None

    _global_options = []

    @classmethod
    def set_global_options(cls, *args: str) -> None:
        """
        set ssh command options
        :param args: options to be applied to all invocations of ssh
        """
        cls._global_options = args

    @property
    def destination(self) -> str:
        """
        :return: host if no username provided, else "username@host" format for SSH destination
        """
        return f"{self.username}@{self.host}" if self.username else self.host

    async def _remote_execute(self, command: str,
                              stdout, stderr, stdin=None,
                              env=None,
                              cwd=None,
                              prefix_cmd: Optional[str] = None,
                              auth_key: Optional[bytes] = None,
                              shell: bool = True) -> asyncio.subprocess.Process:
        """
        Execute given command on remote client a la ssh
        :param command: command to execute
        :return: proc
        """
        env = env or {}
        if auth_key is not None:
            env["AUTH_TOKEN_STDIN"] = '1'
        if env:
            for key, value in env.items():
                command = f"{key}=\"{value}\" {command}"
        if user_output.verbose:
            env["PTMPROC_VERBOSE"] = '1'
        if cwd is not None:
            command = f"cd {str(cwd)} && {prefix_cmd or ''} {command}"
        command = command.replace('"', '\\\"')
        full_cmd = f"ssh {self.destination} {' '.join(self._global_options)} \"{command}\""
        debug_print(f"Executing command '{full_cmd}'")
        if shell:
            result = await asyncio.subprocess.create_subprocess_shell(
                full_cmd,
                stdout=stdout,
                stderr=stderr,
                stdin=stdin,
            )
            assert result
            return result
        else:
            return await asyncio.subprocess.create_subprocess_exec(
                'ssh', self.destination, *self._global_options, f"\"command\"",
                stdout=stdout,
                stderr=stderr,
                stdin=stdin,
            )

    async def mkdir(self, remote_path: Path, exists_ok=True):
        """
        make directory on remote client
        :param remote_path: path to create
        :param exists_ok: if True, do not raise error if directory already exists
        :raises: CommandExecutionFailure if command fails to execute (non-zero return code)
        """
        options = "-p" if exists_ok is True else ""
        proc = await self._remote_execute(
            f"mkdir {options} {str(remote_path)}",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT
        )
        rc = await asyncio.wait_for(proc.wait(), timeout=5)
        if rc != SUCCESS:
            stdout = await proc.stdout.read()
            if not exists_ok or b'File exists' not in stdout:
                raise CommandExecutionFailure(f"mkdir -p {remote_path}:\n\n{stdout}", rc)

    async def rmdir(self, remote_path: Path):
        """
        Remove directory from remote client
        :param remote_path: which directory
        :raises: CommandExecutionFailure if command execution fails on remote client
        """
        proc = await self._remote_execute(
            f"rm -r {remote_path}",
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        rc = await asyncio.wait_for(proc.wait(), timeout=10)
        if rc != SUCCESS:
            raise CommandExecutionFailure(f"rm -rf {remote_path}", rc)

    async def remove(self, remote_path: Path):
        """
        Remove path from remote client

        :param remote_path: path to file to remove
        :raises: CommandExecutionFailure if command execution fails on remote client
        """
        proc = await self._remote_execute(
            f"rm {remote_path}",
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        rc = await asyncio.wait_for(proc.wait(), timeout=10)
        if rc != SUCCESS:
            raise CommandExecutionFailure(f"rm -rf {remote_path}", rc)

    async def push(self, local_path: Union[Path, List[Path]], remote_path: Path, timeout: Optional[float] = None):
        """
        Puh file(s) to a remote client
        :param local_path: path to local file or (recursive) directory to copy
        :param remote_path: destination on remote client to copy to
        :param timeout: optional timeout for command execution
        :raises: TimeoutError if failed to execute in time
        :raises: CommandExecutionError if command fails to execute on remote client
        """
        if isinstance(local_path, list):
            file_list = ' '.join([f"\"{str(p)}\"" for p in local_path])
            args = list(self._global_options) +\
                ['-r', '-p', file_list, f"{self.destination}:{str(remote_path)}"]
        elif local_path.is_dir():
            args = list(self._global_options) +\
                ['-r', '-p', str(local_path), f"{self.destination}:{str(remote_path)}"]
        else:
            args = list(self._global_options) +\
                ['-p', str(local_path), f"{self.destination}:{str(remote_path)}"]
        debug_print(f"Executing command 'scp {' '.join(args)}")
        proc = await asyncio.create_subprocess_shell(
            f"scp {' '.join(args)}",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        if timeout:
            rc = await asyncio.wait_for(proc.wait(), timeout=timeout)
        else:
            rc = await proc.wait()
        if rc != SUCCESS:
            stdout = await proc.stdout.read()
            stderr = await proc.stderr.read()
            raise CommandExecutionFailure(f"Copy from {str(local_path)} to {str(remote_path)} [scp {' '.join(args)}] \n\n{stdout}\n\n {stderr}\n\n",
                                          rc=rc)

    async def pull(self, remote_path: Path, local_path: Path, recursive: bool = True, timeout: Optional[float] = None):
        """
        Pull file(s) from remote client

        :param remote_path: file or directory to pull (recurseively) from remote client
        :param local_path: path to copy to local host
        :param recursive: specify True if copying a directory
        :param timeout: optional timeout for command to execute
        :raises: TimeoutError if command fails to execute in time
        :raises: CommendExecutionFailure if command faile to execute on remote host
        """
        if recursive or local_path.is_dir():
            args = ['-p', '-r',  f"{self.destination}:{str(remote_path)}", str(local_path.absolute())]
        else:
            args = ['-p', f"{self.destination}:{str(remote_path)}", str(local_path.absolute())]
        proc = await asyncio.create_subprocess_exec(
            "scp", *self._global_options, *args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        if timeout:
            rc = await asyncio.wait_for(proc.wait(), timeout=timeout)
        else:
            rc = await proc.wait()
        if rc != 0:
            stdout = await proc.stdout.read()
            cmd = "scp " + ' '.join(self._global_options) + ' ' + ' '.join(args)
            msg = f"\n!!! Copy from  {self.destination}:{str(remote_path)} to {str(local_path.absolute())} using\n"\
                  f"{cmd}\n\n{stdout.decode('utf-8')}"
            raise CommandExecutionFailure(msg, rc)

    async def get_remote_platform_info(self):
        """
        :return:  Tuple of system and machine platform attributes of remote client
        """
        proc = await self._remote_execute(
            f"uname -s && uname -p",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )
        stdout, _ = await asyncio.wait_for(proc.communicate(), timeout=5)
        if proc.returncode != 0:
            raise CommandExecutionFailure("Getting platform info through uname", proc.returncode)
        system, machine = stdout.strip().splitlines()
        if isinstance(system, bytes):
            system = system.decode('utf-8')
            machine = machine.decode('utf-8')
        return system, machine

    async def install_packages(
            self,  *pkgs,
            venv: Path,
            cwd: Optional[Path] = None,
            stdout: Optional[Union[int, TextIO]] = None,
            stderr: Optional[Union[int, TextIO]] = sys.stderr):
        remote_py_executable = venv / 'bin' / 'python3'
        cmd = f"{str(remote_py_executable)} -m pip install --upgrade {' '.join(pkgs)}"
        proc = await self._remote_execute(
            cmd,
            stdout=stdout,
            stderr=stderr,
            cwd=str(cwd) if cwd is not None else None
        )
        stdout, _ = await proc.communicate()
        assert proc.returncode is not None
        if proc.returncode != 0:
            os.write(sys.stderr.fileno(), f"Command {cmd} failed to install on remote".encode('utf-8'))
            raise CommandExecutionFailure(cmd, proc.returncode)

    async def install(self,
                      venv: Path,
                      remote_root: Path,
                      requirements_path: str,
                      cwd: Optional[Path] = None,
                      stdout: Optional[Union[int, TextIO]] = None,
                      stderr: Optional[Union[int, TextIO]] = sys.stderr):
        """
        install Python requirements  on remote client

        :param cwd: directory to run from on remote host
        :param venv: path to python virtual environment on remote host
        :param remote_root: root directory on remote client
        :param requirements_path: path, relative to remote_root, where requirments file is to be found
        :param stdout: as per subprocess.Popen
        :param stderr: as per subprocess.Popen
        :raises: CommandExecutionFailure if install failes on remote client
        """
        remote_py_executable = venv / 'bin' / 'python3'
        cmd = f"{str(remote_py_executable)} -m pip install --upgrade -r {str(remote_root / requirements_path)}"
        proc = await self._remote_execute(
            cmd,
            stdout=stdout,
            stderr=stderr,
            cwd=str(cwd) if cwd is not None else None
        )
        stdout, _ = await proc.communicate()
        assert proc.returncode is not None
        if proc.returncode != 0:
            os.write(sys.stderr.fileno(), f"Command {cmd} failed to install on remote".encode('utf-8'))
            raise CommandExecutionFailure(cmd, proc.returncode)

    async def monitor_remote_execution(self, command, *args,
                                       timeout: Optional[float] = None,
                                       cwd: Optional[Path] = None,
                                       stdin: Optional[Union[TextIO, int]] = None,
                                       env: Optional[Dict[str, str]] = None,
                                       success_code: Optional[Callable[[int], bool]] = lambda x: x == 0)\
            -> AsyncIterator[str]:
        """
        Monitor remote execution of a command asynchronously (asyncio)

        :param command: command to execute on remote client
        :param args: args to the command
        :param timeout: optional timeout for command
        :param cwd: directory on remote client in which to execute
        :param success_code: None if not to check return code, otherwise lambda taking return code an returning
           True is success or False otherwise (defaults to 0 being success code)
        :param stdin: stream to use for stdin (as in subprocess.Popen)
        :param env: optional dict of environment vars to apply

        :return: AsyncIterator of string providing line-by-line output from command
        :raises: TimeoutError if command fails to finish in time
        :raises: CommandExecutionFailure if command return non-success error code
        """
        start_time = time.monotonic()

        def getlines(data: str) -> Tuple[str, str]:
            """
            Split data into lines, but treat any last remnant as a buffer for client to combine with
            a next batch of data before declaring a full line

            :param data: string to process
            :return: a tuple of lines contained in a string and a buffer of data leftover at end
            """
            if '\n' in data:
                *lines, buffer = data.splitlines()
                if data.endswith('\n'):
                    lines.append(buffer)
                    return lines, ""
                else:
                    return lines, buffer
            else:
                return "", data
        proc = await self.launch_remote_command(command, *args,
                                                stdin=stdin,
                                                stdout=asyncio.subprocess.PIPE,
                                                stderr=asyncio.subprocess.STDOUT, cwd=cwd,
                                                env=env or os.environ)
        buffer_stdout = ""
        time_left = timeout or 0.0
        while time_left > 0.0:
            data = (await asyncio.wait_for(proc.stdout.read(10*1024), timeout=time_left)).decode('utf-8')
            lines, buffer_stdout = getlines(buffer_stdout + data)
            for line in lines:
                yield line
            time_left = timeout - time.monotonic() + start_time
            if not data:
                break
        if timeout is not None and time_left <= 0.0:
            with suppress(OSError):
                proc.terminate()
            with suppress(TimeoutError):
                await asyncio.wait_for(proc.wait(), timeout=5)
            if proc.returncode is None:
                with suppress(OSError):
                    proc.kill()
            raise TimeoutError(f"timeout waiting for ssh of command {command}")
        if buffer_stdout:
            yield buffer_stdout
        await asyncio.wait_for(proc.wait(), timeout=timeout)
        if success_code and not success_code(proc.returncode):
            logging.error(f"process '{command} {' '.join(args)}' failed with code {proc.returncode}")
            raise CommandExecutionFailure(command, proc.returncode)

    @asynccontextmanager
    async def mkdtemp(self) -> Path:
        """
        :return: temporary directory created on remote client
        """
        proc = await self._remote_execute(
            f"mktemp -d --tmpdir ptmproc-tmp.XXXXXXXXXX",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            shell=True
        )
        await asyncio.wait_for(proc.wait(), timeout=5)
        if proc.returncode != 0:
            stdout = await proc.stdout.read()
            stderr = await proc.stderr.read()
            raise SystemError(f"Failed to create tmp dir on remote client: {stdout}\n  {stderr}")
        # noinspection SpellCheckingInspection
        tmpdir = (await proc.stdout.read()).decode('utf-8').strip()
        yield Path(tmpdir)
        await self.rmdir(Path(tmpdir))

    async def launch_remote_command(self,  command, *args,
                                    stdin: Union[TextIO, int],
                                    stdout: Union[TextIO, int],
                                    stderr: Union[TextIO, int],
                                    cwd: Optional[Path] = None,
                                    env: Optional[Dict[str, str]] = None,
                                    prefix_cmd: Optional[str] = None,
                                    auth_key: Optional[bytes] = None,
                                    shell:bool = True) -> asyncio.subprocess.Process:
        """
        execute command on remote host

        :param command: command to execute
        :param prefix_cmd: an optional prefix to insert before command in shell command string
        :param args: args to command
        :param cwd: optional directory on remote host to execute under
        :param stdin: stream to use for stdin (as in subprocess.Popen)
        :param stdout: stream to use for stdout (as in subprocess.Popen)
        :param stderr: stream to use for stderr (as in subprocess.Popen)
        :param env: optional dictionary of environment variables for command
        :param auth_key: authentication key for multiprocessing
        :return: Process created (completed process if timeout is specified)
        :raises: TimeoutError if command does not execute in time (if timeout is specified)
        """
        args = [shlex.quote(arg) if '*' not in arg else arg for arg in args]
        full_command = command + ' ' + ' '.join(args)
        result = await self._remote_execute(
            full_command,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=env,
            cwd=cwd,
            prefix_cmd=prefix_cmd,
            auth_key=auth_key,
            shell=shell
        )
        assert result
        return result

    async def execute_remote_cmd(self,  command, *args,
                                 stdout: Union[int, TextIO] = sys.stdout,
                                 stderr: Union[int, TextIO] = sys.stderr,
                                 stdin: Optional[TextIO] = None,
                                 timeout: Optional[float] = None,
                                 cwd: Optional[Path] = None,
                                 prefix_cmd: Optional[str] = None,
                                 env: Optional[Dict[str, str]] = None,
                                 auth_key: Optional[bytes] = None,
                                 shell: bool = True,
                                 ) -> asyncio.subprocess.Process:
        """
        execute command on remote host

        :param command: command to execute
        :param prefix_cmd: an optional prefix to insert before command in shell command string
        :param args: args to command
        :param timeout: optional timeout for command execution;  if None, function will return Process created
           without waiting for completion
        :param cwd: optional directory on remote host to execute under
        :param stdin: stream to use for stdin (as in subprocess.Popen)
        :param stdout: stream to use for stdout (as in subprocess.Popen)
        :param stderr: stream to use for stderr (as in subprocess.Popen)
        :param env: optional dictionary of environment variables for command
        :param auth_key: authentication key for multiprocessing
        :return: Process created (completed process if timeout is specified)
        :raises: TimeoutError if command does not execute in time (if timeout is specified)
        """
        proc = await self.launch_remote_command(command, *args,
                                                stdin=asyncio.subprocess.PIPE if auth_key is not None else stdin,
                                                stdout=stdout,
                                                stderr=stderr,
                                                cwd=cwd,
                                                auth_key=auth_key,
                                                prefix_cmd=prefix_cmd,
                                                shell=shell,
                                                env=env)
        if auth_key is not None:
            import binascii
            auth_key = binascii.b2a_hex(auth_key)
            proc.stdin.write(auth_key + b'\n')
            proc.stdin.close()
        await asyncio.wait_for(proc.wait(), timeout=timeout)
        return proc


@asynccontextmanager
async def remote_root_context(project_name: str, ssh_client: SSHClient, remote_root: Optional[Path]) -> \
        AsyncContextManager[Tuple[Path, Path]]:
    """
    Provide a context representing a directory on the remote host from which to run.  If no explicit
    path is provided as a remote root, a temporary directory is created on remote host and removed
    when exiting this context.  (Otherwise context manager has no real effect)

    :param project_name: name of project, used in createing a cache dir if available to remote system
    :param ssh_client: ssh client to use in creating a temp dir if necessary
    :param remote_root: an explicit location on the remote host (will not be remove on exit of context)
    :return: the remote root created (or reflects the remote root provided explicitly)
    """
    allow_cache = os.environ.get('PTMPROC_DISABLE_CACHE', False) not in ('1', 'True', 'true', 'TRUE')
    if allow_cache:
        proc = await ssh_client.execute_remote_cmd('echo \${HOME}', stdout=asyncio.subprocess.PIPE, shell=True)
        stdout_text = (await proc.stdout.read()).strip().decode('utf-8')
        if proc.returncode != 0 or not stdout_text:
            var_cache_path = Path('/var') / 'cache' / 'ptmproc'
        else:
            var_cache_path = Path(stdout_text) / '.ptmproc' / 'cache'
        # noinspection PyBroadException
        try:
            await ssh_client.mkdir(var_cache_path / project_name, exists_ok=True)
            debug_print(f"Using {var_cache_path / project_name} to cache python venv")
            venv_root = var_cache_path / project_name
        except Exception:
            venv_root = None
    else:
        venv_root = None
    if remote_root:
        yield remote_root, venv_root or remote_root
    else:
        async with ssh_client.mkdtemp() as root:
            debug_print(f"Using dir {venv_root} to set up python venv and {root} as temp staging dir")
            yield root, venv_root or root



import os
from pathlib import Path

import pytest

from pytest_mproc.ptmproc_data import ProjectConfig, RemoteHostConfig
from pytest_mproc.remote.bundle import Bundle


class TestProjectConfig:
    ROOT = Path(__file__).parent / "resources"

    def test_from_file(self):
        project = ProjectConfig.from_file(self.ROOT / "project.cfg")
        assert project.src_paths == [Path("./src1"), Path("src2")]
        assert project.test_files == [
            Path("project_tests/*.py"),
            Path("./requirements1.txt"),
            Path("./requirements2.txt"),
            Path("./resources1/*"),
            Path("resources2/*"),
        ]

    def test_from_file_invalid_json(self):
        with pytest.raises(pytest.UsageError) as e:
            ProjectConfig.from_file(self.ROOT / "project_invalid_json.cfg")
        assert "Invalid json" in str(e)

    @pytest.mark.parametrize("key", ['src_paths', 'test_files'])
    def test_from_file_no_such_dir_or_file(self, chdir, key):
        chdir.chdir(self.ROOT)
        with pytest.raises(Exception) as e:
            project_config = ProjectConfig.from_file((self.ROOT / f"project_{key}_no_such_dir.cfg"))
            Bundle.create(self.ROOT, project_config)
        if isinstance(e.value, pytest.UsageError):
            assert "is not a path" in str(e)
        elif isinstance(e.value, ValueError):
            assert "did not match" in str(e)
        else:
            raise e.value



class TestRemoteHostConfig:
    def test_from_list(self):
        config = RemoteHostConfig.from_list(
            ['129.1.2.10;device=DEVICE1;arg2=VALUE',
             '129.1.2.11;device=DEVICE2']
        )
        configs = {config.remote_host: config for config in config}
        assert '129.1.2.10' in configs
        assert '129.1.2.11' in configs
        assert configs['129.1.2.10'].remote_host == '129.1.2.10'
        assert configs['129.1.2.10'].arguments == {'device': "DEVICE1",
                                                   'arg2': "VALUE"}
        assert configs['129.1.2.11'].remote_host == '129.1.2.11'
        assert configs['129.1.2.11'].arguments == {'device': "DEVICE2"}

    def test_from_list_bad_string(self):
        with pytest.raises(pytest.UsageError):
            RemoteHostConfig.from_list(
                ['129.1.2.10;device=DEVICE1;arg',
                 '129.1.2.11;device=DEVICE2']
            )

    @pytest.mark.asyncio
    async def test_from_uri_string(self):
        os.chdir(Path(__file__).parent)
        string = "file://./resources/remote_host.cfg"
        configs = {}
        it = RemoteHostConfig.from_uri_string(string)
        async for config in it:
            configs[config.remote_host] = config
        assert '192.1.3.64' in configs
        assert '192.1.3.65' in configs
        assert configs['192.1.3.64'].remote_host == '192.1.3.64'
        assert configs['192.1.3.64'].arguments == {"device": "DEVICE1"}
        assert configs['192.1.3.65'].remote_host == '192.1.3.65'
        assert configs['192.1.3.65'].arguments == {"device": "DEVICE2", 'arg': "VAL"}



import asyncio
import logging
import os
import platform
from glob import glob
from multiprocessing import Queue

import binascii
import sys
import tempfile
import zipfile
from contextlib import suppress
from pathlib import Path
from threading import RLock
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from pytest_mproc import user_output, get_auth_key, Constants
from pytest_mproc.fixtures import Node
from pytest_mproc.ptmproc_data import RemoteHostConfig, ProjectConfig
from pytest_mproc.remote.ssh import SSHClient, CommandExecutionFailure, remote_root_context
from pytest_mproc.user_output import debug_print, always_print

_root = os.path.dirname(__file__)
FIFTEEN_MINUTES = 15*60
artifacts_root = os.environ.get('PTMPROC_HOST_ARTIFACTS_DIR', './artifacts')
lock = RLock()


class Bundle:

    # noinspection SpellCheckingInspection
    CACHE_DIR = Path(os.path.expanduser('~')).absolute() / ".ptmproc"

    def __init__(self, root_dir: Path,
                 project_name: str,
                 project_root: Path,
                 artifacts_path: Path,
                 test_files: List[Path],
                 system_executable: str,
                 prepare_script: Optional[Path] = None,
                 finalize_script: Optional[Path] = None,
                 env: Optional[Dict[str, str]] = None,
                 relative_run_dir: Optional[Path] = None):
        """
        create a bundle for remote host deployment

        :param root_dir: a delicious carbonated beverage (and also the root where bundle files are created)
        :param test_files: location of test files, resources files, requirements files to upload to run tests
        :param system_executable: location of python executable on remote host, or None for default
        :param prepare_script: Optional path to script to run pre-test-execution
        :param finalize_script:: Optional path to run post-test-execution
        :param relative_run_dir: working directory, relative to root, where pytest should be executed; root dir if
           not specified
        """
        self._project_name = project_name
        self._artifacts_path = artifacts_path
        if not root_dir.exists():
            raise FileNotFoundError(f"'{root_dir}' does not exist")
        if relative_run_dir is not None and relative_run_dir.is_absolute():
            raise ValueError("Relative run dir was provided as an absolute path when creating bundle")
        absolutes = [p for p in test_files if p.is_absolute()]
        if absolutes:
            raise ValueError(f"Files or wildcard patterns {absolutes} in project config file must be relative paths "
                             "to the project config location")
        self._zip_path = root_dir / "bundle.zip"
        if os.path.exists(self._zip_path):
            raise FileExistsError(f"File {self._zip_path} already exists")
        self._root_dir = Path(root_dir)
        self._prepare_script = Path(prepare_script) if prepare_script else None
        self._finalize_script = Path(finalize_script) if finalize_script else None
        self._remote_executable = system_executable or Path(sys.executable).name
        self._sources: List[Path] = []
        self._site_pkgs = tempfile.TemporaryDirectory()
        self._resources: List[Tuple[Path, Path]] = []
        self._requirements_path = None
        self._test_files = []
        self._relative_run_path = relative_run_dir or Path(".")
        for path in test_files:
            matched = glob(str(project_root / path))
            if not matched:
                raise ValueError(
                    f"files or wildcard pattern '{str(project_root.absolute() / path)}' did not match any files")
            self._test_files.extend([Path(p).relative_to(project_root) for p in matched])
        if any([not (project_root / p).exists() for p in self._test_files]):
            raise FileNotFoundError(f"One or more files in '{test_files}' do not exist in {project_root}")
        if Path('requirements.txt') in self._test_files:
            self._requirements_path = 'requirements.txt'
        self._system = platform.system()
        self._machine = platform.machine()
        self._env = env or {}

    @staticmethod
    def remote_run_dir(remote_root: Path) -> Path:
        """
        :return: Path on remote host where test code and resources are located for the run, given the root dir
           of installation of a bundle
        """
        return remote_root / "run"

    @property
    def prepare_script(self):
        return self._prepare_script

    @property
    def finalize_script(self):
        return self._finalize_script

    def __enter__(self) -> "Bundle":
        self._site_pkgs.__enter__()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._site_pkgs.__exit__(exc_type, exc_val, exc_tb)

    @property
    def zip_path(self):
        return self._zip_path

    @property
    def tests_files(self):
        return self._test_files

    def add_file(self, path: Union[Path, str], relative_path: Union[Path, str]) -> None:
        """
        add given file to bundle
        :param path: path to the file to add
        :param relative_path: path relative remote root directory
        """
        path = Path(path) if isinstance(path, str) else path
        relative_path = Path(relative_path) if isinstance(relative_path, str) else relative_path
        if relative_path.is_absolute():
            raise Exception("relative path to add must be relative, not absolute")
        if not path.exists() or not path.is_file():
            raise FileExistsError(f"File {path} does not exist or is not a file")
        self._resources.append((path, relative_path))

    @classmethod
    def create(cls,
               root_dir: Path,
               project_config: ProjectConfig,
               system_executable: Optional[str] = None,
               ) -> "Bundle":
        """
        Create a bundle
        :param root_dir: where artifacts are to be created
        :param project_config: project configuration defining contents of bundle
        :param system_executable: path ON REMOTE HOST of system python executable to use
        :return: created Bundle
        """
        try:
            relative_run_dir = Path(os.getcwd()).relative_to(project_config.project_root.absolute())
        except ValueError:
            raise ValueError("You must execute pytest in the directory containing the project config, "
                             f"{str(project_config.project_root)} or a subdirectory thereof")
        bundle = Bundle(root_dir=root_dir,
                        project_name=project_config.project_name,
                        artifacts_path=project_config.artifcats_path,
                        project_root=project_config.project_root,
                        test_files=project_config.test_files,
                        prepare_script=project_config.prepare_script,
                        finalize_script=project_config.finalize_script,
                        relative_run_dir=relative_run_dir,
                        system_executable=system_executable)
        bundle._sources.extend(project_config.src_paths)
        os.write(sys.stderr.fileno(), f"\n\n>>> Building bundle.  This may take a little while...\n\n".encode('utf-8'))
        # site_pkgs = env.set_up_local(project_config=project_config, cache_dir=cls.CACHE_DIR)
        # bundle._sources.append(site_pkgs)
        always_print(">>> Zipping contents for remote worker...")
        with zipfile.ZipFile(bundle.zip_path, mode='w') as zfile:
            for path in bundle._test_files:
                zfile.write(project_config.project_root / path, f"run{os.sep}{str(path)}")
            if bundle.prepare_script:
                zfile.write(project_config.project_root / bundle.prepare_script, 'prepare')
            if bundle.finalize_script:
                zfile.write(project_config.project_root / bundle.finalize_script, 'finalize')
            with tempfile.NamedTemporaryFile(mode='a+') as out:
                out.write(bundle._system)
                out.write(bundle._machine)
                out.flush()
                zfile.write(out.name, "host_ptmproc_platform")
            for source_path in bundle._sources:
                for f in (project_config.project_root / source_path).glob("**/*"):
                    if f.is_file() and '__pycache__' not in f.name:
                        zfile.write(f, f"site-packages/{f.relative_to(project_config.project_root / source_path)}")
        return bundle

    async def setup_remote_venv(self, ssh_client: SSHClient, remote_venv: Path) -> Path:
        """
        setup the remote base virtual environment (Python + pytest + pytest_mproc)
        """
        stdout = sys.stdout if user_output.verbose else asyncio.subprocess.DEVNULL
        proc = await ssh_client.execute_remote_cmd(
            'ls', '-d', str(remote_venv / 'bin' / 'python3'), stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL
        )

        if proc.returncode != 0:
            always_print(f"Installing python virtual environment to {remote_venv}")
            await ssh_client.execute_remote_cmd(
                self._remote_executable, '-m', 'venv', str(remote_venv),
                stdout=stdout,
                stderr=sys.stderr
            )
            await ssh_client.install_packages(
                'pytest', 'pytest_mproc',
                venv = remote_venv,
                cwd = remote_venv.parent,
                stdout = stdout,
                stderr = sys.stderr)

        return remote_venv

    async def deploy(self, ssh_client: SSHClient,
                     remote_venv: Path, remote_root: Path, timeout: Optional[float] = None) -> None:
        """
        Deploy this bundle to a remote ssh client
        :param ssh_client: where to deploy
        :param timeout: raise TimeoutError if deployment takes too long
        :param remote_root: remote path to where bundle is deployed
        """
        always_print("Deploying bundle to remote worker %s...", ssh_client.host)
        stdout = sys.stdout if user_output.verbose else asyncio.subprocess.DEVNULL
        try:
            await ssh_client.mkdir(self.remote_run_dir(remote_root), exists_ok=True)
            debug_print(f">>> Pushing contents to remote worker {ssh_client.host}...")
            await ssh_client.push(self._zip_path, remote_root / self._zip_path.name)
            debug_print(f">>> Unpacking tests on remote worker {ssh_client.host}...")
            # noinspection PyProtectedMember
            proc = await ssh_client._remote_execute(
                f"cd {str(remote_root)} && unzip -o {str(self._zip_path.name)}",
                stdout=stdout,
                stderr=sys.stderr
            )
            await asyncio.wait_for(proc.wait(), timeout=timeout)
            if proc.returncode != 0:
                text = ('\n' + (await proc.stdout.read()).decode('utf-8')) if proc.stdout else ""
                raise CommandExecutionFailure(f"Failed to unzip tests on remote client {text}", proc.returncode)
            with suppress(Exception):
                await ssh_client.remove(remote_root / self._zip_path.name)
            if proc.returncode != 0:
                text = ('\n' + (await proc.stdout.read()).decode('utf-8')) if proc.stdout else ""
                raise CommandExecutionFailure(f"Failed to unzip requirements on remote client {text}", proc.returncode)
            if self._requirements_path:
                always_print(f"Installing requirements on remote worker {ssh_client.host}...")
                await ssh_client.install(
                    venv=remote_venv,
                    remote_root=remote_root,
                    requirements_path=f'run{os.sep}requirements.txt',
                    cwd=remote_root,
                    stdout=stdout,
                    stderr=sys.stderr
                )
            await ssh_client.mkdir(remote_root / "run" / self._relative_run_path, exists_ok=True)
        except Exception as e:
            always_print(f"!!! Failed to deploy to {ssh_client.host}: {e}")
            raise

    async def delegate(self, ssh_client: SSHClient,
                       delegation_port: int,
                       remote_root: Path, remote_venv: Path)\
            -> Tuple[asyncio.subprocess.Process, int]:
        """
        Launch a process to delegate main manager to a worker node
        This is done when the main node is blcoekd (by firewall) from receiving incoming data on a port

        :return: created process
        """
        root = Path(__file__).parent.parent.parent
        await ssh_client.mkdir(remote_root / 'site-packages' / 'pytest_mproc' / 'remote', exists_ok=True)
        await ssh_client.push(root / 'pytest_mproc' / '*.py', remote_root / 'site-packages' / 'pytest_mproc' / '.')
        await ssh_client.push(root / 'pytest_mproc' / 'remote' / '*.py', remote_root / 'site-packages' / 'pytest_mproc' / 'remote' / '.')
        python = str(remote_venv / 'bin' / 'python3')
        env = {'PYTHONPATH': str(remote_root / 'site-packages'),
               'AUTH_TOKEN_STDIN': '1'}
        delegation_proc = await ssh_client.launch_remote_command(
            python, '-m', 'pytest_mproc.orchestration', ssh_client.host,
            str(delegation_port), self._project_name,
            env=env,
            stderr=sys.stderr,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
        )
        delegation_proc.stdin.write(binascii.b2a_hex(get_auth_key()) + b'\n')
        delegation_proc.stdin.close()
        port_text = (await delegation_proc.stdout.readline()).decode('utf-8')
        return delegation_proc, int(port_text)

    async def execute_remote_multi(
            self,
            server_info: Tuple[str, int],
            hosts_q: Queue,
            port_q: Queue,
            username: Optional[str] = None,
            password: Optional[str] = None,
            deploy_timeout: Optional[float] = FIFTEEN_MINUTES,
            timeout: Optional[float] = None,
            auth_key: Optional[bytes] = None,
            env: Optional[Dict[str, str]] = None,
            delegation_port: Optional[int] = None,
    ) -> Tuple[Dict[str, asyncio.subprocess.Process], str, asyncio.subprocess.Process]:
        """
        deploy and execute to/on multiple host targets concurrently (async)

        :param username: optional username for login-based ssh
        :param password: optional password for login-based ssh
        :param auth_key: auth token to use in multiprocessing authentication
        :param deploy_timeout: optional timeout if deployment takes too long
        :param env: dict of environment variables to use on execution
        :param timeout: optional overall timeout for execution to complete on all hosts
        :return: Dict of host, proc where host is host process executed on and proc is the completed process
        """
        tasks: List[asyncio.Task] = []
        deployments = {}
        contexts = {}
        deployment_tasks: Dict[str, asyncio.Task] = {}
        delegation_proc: Optional[asyncio.subprocess.Process] = None
        delegation_host: Optional[str] = None
        try:
            env = env.copy() if env else {}
            index = 0
            worker_config: RemoteHostConfig = hosts_q.get()
            server_host, server_port = server_info
            remote_roots: Dict[str, Path] = {}
            remote_venvs: Dict[str, Path] = {}
            while worker_config is not None:
                ssh_client = SSHClient(worker_config.remote_host, username=username, password=password)
                if worker_config.remote_host not in contexts:
                    context = remote_root_context(self._project_name, ssh_client, worker_config.remote_root)
                    contexts[worker_config.remote_host] = context
                    remote_roots[worker_config.remote_host], remote_venv_root = await context.__aenter__()
                    remote_venvs[worker_config.remote_host] = remote_venv_root / 'venv'
                remote_root = remote_roots[worker_config.remote_host]
                remote_venv = remote_venvs[worker_config.remote_host]
                if worker_config.remote_host not in deployments:
                    await self.setup_remote_venv(ssh_client=ssh_client, remote_venv=remote_venv)
                if index == 0 and delegation_port is not None:
                    delegation_host = worker_config.remote_host
                    delegation_proc, server_port = await self.delegate(
                        ssh_client=ssh_client,
                        delegation_port=delegation_port, remote_venv=remote_venv,
                        remote_root=remote_root)
                    port_q.put(server_port)
                    server_host = worker_config.remote_host
                index += 1
                cli_args, cli_env = worker_config.argument_env_list()
                env.update(cli_env)
                if worker_config.remote_host not in deployment_tasks:
                    deployment_tasks[worker_config.remote_host] = asyncio.create_task(self.deploy(
                        ssh_client=ssh_client,
                        remote_root=remote_root,
                        remote_venv=remote_venv,
                        timeout=deploy_timeout
                    ))

                async def execute(worker_config: RemoteHostConfig):
                    host = worker_config.remote_host
                    # we must wait if deployment is not finished
                    # this setup parallelizes multiple host efficiently
                    if deployment_tasks[host]:
                        await deployment_tasks[host]
                        deployment_tasks[host] = False
                    args = _determine_cli_args(worker_config)
                    return await self.execute_remote(
                        *(list(args) + list(cli_args)) + ["--as-worker", f"{server_host}:{server_port}"],
                        worker_id=f"Worker-{index}",
                        env=env,
                        cwd=remote_root / f'Worker-{index}',
                        ssh_client=ssh_client,
                        timeout=timeout,
                        deploy_timeout=deploy_timeout,
                        remote_root=remote_root,
                        auth_key=auth_key,
                        remote_venv=remote_venv,
                        deploy=False
                    )

                task = asyncio.get_event_loop().create_task(execute(worker_config))
                tasks.append(task)
                if worker_config.remote_host not in deployments:
                    deployments[ssh_client.host] = (remote_root, remote_venv)
                worker_config = hosts_q.get()
            results, _ = await asyncio.wait(tasks, timeout=timeout,
                                            return_when=asyncio.ALL_COMPLETED)
            results = [r.result() for r in results]
            self._artifacts_path.mkdir(exist_ok=True)
            with open(self._artifacts_path / "worker_info.txt", 'w') as out_stream:
                for index, r in enumerate(results):
                    if not isinstance(r, Exception):
                        host, proc = r
                        out_stream.write(f"Worker-{index}: host={host}, pid={proc.pid}")
                    else:
                        out_stream.write(f"Worker-{index}: <<see exception for this worker>>")
            procs = {host: proc for host, proc in results if not isinstance(proc, Exception)}
            exceptions = [e for e in results if isinstance(e, Exception)]
            if exceptions:
                for e in exceptions:
                    logging.exception(e, msg=f"Error execution on remote host")
                if not procs:
                    raise Exception(f"Exception(s) during remote execution on all remote hosts : {exceptions}")
            for proc in procs.values():
                if proc.returncode is None:
                    with suppress(OSError):
                        proc.kill()
            return procs, delegation_host, delegation_proc
        finally:
            for context in contexts.values():
                await context.__aexit__(None, None, None)

    async def execute_remote(self, *args,
                             cwd: Path,
                             remote_root: Path,
                             remote_venv: Path,
                             env: Optional[Dict[str, str]] = None,
                             ssh_client: SSHClient,
                             timeout: Optional[float] = None,
                             deploy_timeout: Optional[float] = None,
                             auth_key: Optional[bytes] = None,
                             worker_id: Optional[str] = None,
                             deploy: bool = True
                             ) -> Tuple[str, asyncio.subprocess.Process]:
        """
        Execute tests on remote host
        :param ssh_client: ssh client to use to execute remotely
        :param args: args to pass on cmd line
        :param env: additional environment variables to set (key/value pairs)
        :param auth_key: token to use for multiprocessing authentication
        :param timeout: optional timeout if execution takes too long; if None return proc instance immediately
            with no waiting
        :param remote_root: where on remote host to deploy bundle
        :param remote_venv: location of virtual Python environment to use on host (possibly from cache)
        :param worker_id: Optional identifier to assign to worker
        :return: asyncio.subprocess.Process instance
        """
        if deploy:
            await self.deploy(
                ssh_client=ssh_client,
                remote_root=remote_root,
                remote_venv=remote_venv,
                timeout=deploy_timeout
            )
        command = str(remote_venv / 'bin' / 'python3')
        env.update(self._env)
        if auth_key:
            env['AUTH_TOKEN_STDIN'] = "1"
        # noinspection SpellCheckingInspection
        env['PTMPROC_EXECUTABLE'] = command.split()[0]
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] = str(remote_root / 'site-packages') + ':' + env["PYTHONPATH"]
        else:
            env["PYTHONPATH"] = str(remote_root / 'site-packages')
        if user_output.verbose:
            env["PTMPROC_VERBOSE"] = '1'
        env['PTMPROC_NODE_MGR_PORT'] = str(Node.Manager.PORT)
        if "PTMPROC_BASE_PORT" in os.environ:
            env["PTMPROC_BASE_PORT"] = os.environ["PTMPROC_BASE_PORT"]
        always_print("Executing tests on remote worker %s...", ssh_client.host)
        run_dir = self.remote_run_dir(remote_root)
        await ssh_client.mkdir(cwd, exists_ok=True)
        await ssh_client.execute_remote_cmd("ln", "-sf", str(run_dir / '*'), str(cwd / '.'))
        remote_artifacts_dir = cwd / self._artifacts_path
        await ssh_client.mkdir(remote_artifacts_dir, exists_ok=True)
        cwd = cwd / self._relative_run_path
        pid = None
        proc = None
        stdout = sys.stdout if user_output.verbose else asyncio.subprocess.DEVNULL
        try:
            if self.prepare_script:
                always_print("Preparing environment through prepare script...")
                # noinspection SpellCheckingInspection
                proc = await ssh_client.execute_remote_cmd(
                    f"./prepare | tee \"{remote_artifacts_dir}{os.sep}{self._prepare_script.stem}.log\"",
                    prefix_cmd=f"set -o pipefail; ",
                    cwd=cwd,
                    env=env,
                    timeout=timeout,
                    stdout=stdout,
                    stderr=sys.stderr,
                )
                if proc.returncode != 0:
                    msg = f"!!! Failed to execute prepare script on host [{proc.returncode}]; " \
                          f"see log in artifacts. worker-{worker_id}{os.sep}{self._prepare_script.stem}.log"
                    debug_print(msg)
                    raise SystemError(msg)
            cmd = f"{command} -m pytest {' '.join(args)} -s | "\
                  f"tee {remote_artifacts_dir}{os.sep}pytest_stdout.log"
            debug_print(">>> From %s running %s\n\n %s", cwd, cmd, env)
            await ssh_client.mkdir(remote_artifacts_dir, exists_ok=True)
            env["PTMPROC_WORKER_ARTIFACTS_DIR"] = str(remote_artifacts_dir)
            env["PTMPROC_VERBOSE"] = '1'
            proc = await ssh_client.execute_remote_cmd(
                cmd,
                timeout=timeout,
                env=env,
                prefix_cmd=f"set -o pipefail; ",
                cwd=cwd,
                auth_key=auth_key,
                stdout=sys.stdout,
                stderr=sys.stderr,
                stdin=None if auth_key is None else asyncio.subprocess.PIPE)
            pid = proc.pid
        finally:
            if proc and self._finalize_script:
                always_print("Finalizing worker through finalize script on client %s...", ssh_client.host)
                with suppress(Exception):
                    # noinspection SpellCheckingInspection
                    finalize_proc = await ssh_client.execute_remote_cmd(
                        f"./finalize  | tee \"{remote_artifacts_dir}{os.sep}]{self._finalize_script.stem}.log\"",
                        cwd=cwd,
                        prefix_cmd=f"set -o pipefail; ",
                        env=env,
                        timeout=timeout,
                        stdout=stdout,
                        stderr=sys.stderr,
                    )
                    if finalize_proc.returncode:
                        debug_print(f"!!! WARNING: finalize script failed! ")
            worker_id = worker_id or (f"{ssh_client.host}-{pid}" if pid else ssh_client.host)
            destination_dir = remote_artifacts_dir.absolute()
            try:
                if proc:
                    destination_dir.mkdir(exist_ok=True, parents=True)
                    await ssh_client.mkdir(remote_artifacts_dir, exists_ok=True)
                    await ssh_client.execute_remote_cmd(cwd=remote_artifacts_dir,
                                                        command=f"zip -r {str(remote_artifacts_dir / 'artifacts.zip')} *",
                                                        stdout=stdout,
                                                        stderr=sys.stderr)
                    await ssh_client.pull(remote_artifacts_dir / "artifacts.zip", destination_dir / "artifacts.zip",
                                          recursive=False)
            except CommandExecutionFailure as e:
                os.write(sys.stderr.fileno(),
                         f"\n\n!!! Failed to pull {str(remote_artifacts_dir)} to {destination_dir}:\n"
                         f"    {e}\n\n".encode('utf-8'))
        return ssh_client.host, proc

    async def monitor_remote_execution(self, host: str, *args: str,
                                       username: Optional[str] = None,
                                       password: Optional[str] = None,
                                       deploy_timeout: Optional[float] = FIFTEEN_MINUTES,
                                       timeout: Optional[float] = None,
                                       remote_root: Optional[Path] = None,
                                       env: Optional[Dict[str, str]] = None,
                                       ) -> AsyncGenerator[str, str]:
        ssh_client = SSHClient(host, username, password)
        env = env or {}
        async with remote_root_context(ssh_client, remote_root) as (remote_root, remote_venv_root):
            if "PYTHONPATH" in env:
                env["PYTHONPATH"] = str(remote_root / 'site-packages') + ':' + env["PYTHONPATH"]
            else:
                env["PYTHONPATH"] = str(remote_root / 'site-packages')
            venv = await self.deploy(ssh_client, remote_root=remote_root, remote_venv=remote_venv_root / 'venv',
                                     timeout=deploy_timeout)
            command = venv / 'bin' / 'python3'
            async for line in ssh_client.monitor_remote_execution(
                    command, '-m', 'pytest', *args, timeout=timeout, env=env,
                    cwd=self.remote_run_dir(remote_root) / self._relative_run_path,
            ):
                yield line


def _determine_cli_args(worker_config: RemoteHostConfig):
    args = list(sys.argv[1:])  # copy of
    # remove pytest_mproc cli args to pass to client (aka additional non-pytest_mproc args)
    for arg in sys.argv[1:]:
        typ = Constants.ptmproc_args.get(arg)
        if not typ:
            continue
        if arg == "--cores":
            if "cores" in worker_config.arguments:
                try:
                    index = args.index("--cores")
                    args[index + 1] = str(worker_config.arguments['cores'])
                except ValueError:
                    args += ['--cores', str(worker_config.arguments['cores'])]
        elif typ in (bool,) and arg in args:
            args.remove(arg)
        elif arg in args:
            index = args.index(arg)
            if index + 1 < len(args):
                args.remove(args[index + 1])
            args.remove(arg)
    if "--cores" not in args:
        args += ["--cores", "1"]
    return args


"""
This file contains some standard pytest_* plugin hooks to implement multiprocessing runs
"""
import os
import sys

import _pytest.terminal

from pytest_mproc import worker, user_output, Constants
from pytest_mproc.user_output import debug_print

import getpass
import inspect
import shutil
import tempfile
from contextlib import contextmanager, suppress
from pathlib import Path
from traceback import format_exc
from typing import Callable, Optional, Iterable, Union, Type

from pytest_mproc.ptmproc_data import (
    PytestMprocConfig,
    RemoteHostConfig,
    ProjectConfig,
    PytestMprocRuntime,
)


import pytest
from pytest_mproc import find_free_port, fixtures

import socket

from multiprocessing import cpu_count
from pytest_mproc.utils import is_degraded, BasicReporter


DEPLOY_TIMEOUT = None


# noinspection PyBroadException
def _get_ip_addr():
    try:
        return socket.gethostbyname(socket.gethostname())
    except Exception:
        print(">>> Cannot get ip address of host.  Return 127.0.0.1 (localhost)")
        return "127.0.0.1"


def parse_numprocesses(s: str) -> int:
    """
    A little bit of processing to get number of parallel processes to use (since "auto" can be used to represent
    # of cores on machine)
    :param s: text to process
    :return: number of parallel worker processes to use
    """
    try:
        if s.startswith("auto"):
            if '*' in s:
                multiplication_factor = int(s.rsplit('*', 1)[-1])
            elif s == "auto":
                multiplication_factor = 1
            else:
                raise Exception("Error: --cores argument must be an integer value or auto or auto*<int factor>")
            return cpu_count() * multiplication_factor
        else:
            return int(s)
    except ValueError:
        raise Exception("Error: --cores argument must be an integer value or \"auto\" or \"auto*<int factor>\"")


# noinspection PyProtectedMember
def _add_option(group, name: str, dest: str, action: str,
                typ: Union[Type[int], Type[str], Type[bool], Callable[[str], int]], help_text: str) -> None:
    """
    internal add option function to allow us to keep track of pytest-specific options from command line
    """
    from pytest_mproc.main import Orchestrator
    from pytest_mproc.orchestration import OrchestrationManager
    # noinspection PyProtectedMember
    if typ == bool:
        group._addoption(
            name,
            dest=dest,
            action=action,
            help=help_text,
        )
    else:
        group._addoption(
            name,
            dest=dest,
            action=action,
            type=typ,
            help=help_text,
        )
    Constants.set(name, typ)


@pytest.mark.tryfirst
def pytest_addoption(parser):
    """
    add options to given parser for this plugin
    """
    group = parser.getgroup("pytest_mproc", "better distributed testing through multiprocessing")
    _add_option(
        group,
        "--cores",
        dest="mproc_numcores",
        action="store",
        typ=parse_numprocesses,
        help_text="you can use 'auto' here to set to the number of  CPU cores on host system",
    )
    _add_option(
        group,
        "--disable-mproc",
        dest="mproc_disabled",
        action="store",
        typ=bool,
        help_text="disable any parallel mproc testing, overriding all other mproc arguments",
    )
    _add_option(
        group,
        "--max-simultaneous-connections",
        dest="mproc_max_simultaneous_connections",
        action="store",
        typ=int,
        help_text="max # of connections allowed at one time to main process, to prevent deadlock from overload"
    )
    _add_option(
        group,
        "--as-main",
        dest="mproc_server_uri",
        action="store",
        typ=str,
        help_text="port on which you wish to run server (for multi-host runs only)"
    )
    _add_option(
        group,
        "--remote-worker",
        dest="mproc_remote_clients",
        action="append",
        typ=str,
        help_text="repeatable option to add remote client for execution. string is of "
        "form '<host>[:port][;<key>=<value>]*, "
        "where key, value pairs are translated into command line argument to the host in form '--<key> <value>'"
    )
    if '--as-main' in sys.argv or '--remote-worker' in sys.argv:
        _add_option(
            group,
            "--project-structure",
            dest="project_structure_path",
            action="store",
            typ=str,
            help_text="path to project structure (if server)"
        )
        _add_option(
            group,
            "--remote-sys-executable",
            dest="mproc_remote_sys_executable",
            typ=str,
            help_text="common path on remote hosts to python executable to use",
            action="store"
        )
    else:
        _add_option(
            group,
            "--as-worker",
            dest="mproc_client_connect",
            action="store",
            typ=str,
            help_text="host:port specification of master node to connect to as client"
        )
    _add_option(
        group,
        "--connection-timeout",
        dest="mproc_connection_timeout",
        action="store",
        typ=int,
        help_text="wait this many seconds on connection of client before timing out"
    )
    _add_option(
        group,
        '--ptmproc-verbose',
        dest="mproc_verbose",
        action="store_true",
        typ=bool,
        help_text="output messages when connecting and executing tasks"
    )


def pytest_cmdline_main(config):
    """
    Called before "true" main routine.  This is to set up config values well ahead of time
    for things like pytest-cov that needs to know we are running distributed

    Mostly taken from other implementations (such as xdist)
    """
    mproc_max_connections = getattr(config.option, "mproc_max_simultaneous_connections", 24)
    mproc_connection_timeout = getattr(config.option, "mproc_connection_timeout", None)
    mproc_disabled = getattr(config.option, "mproc_disabled")
    mproc_num_cores = getattr(config.option, "mproc_numcores", None)
    user_output.verbose = config.option.mproc_verbose
    if "PYTEST_WORKER" in os.environ:
        config.option.ptmproc_config = PytestMprocConfig(num_cores=1)
        config.ptmproc_runtime = None
        worker.pytest_cmdline_main(config)
        assert config.option.worker
        return
    else:
        config.option.worker = None
        if config.option.mproc_remote_clients and not config.option.mproc_server_uri:
            config.option.mproc_server_uri = "delegated://"
    if not hasattr(config, "ptmproc_runtime"):
        config.ptmproc_runtime = None

    config.option.ptmproc_config = PytestMprocConfig()
    if mproc_connection_timeout:
        config.option.ptmproc_config.connection_timeout = mproc_connection_timeout
    reporter = BasicReporter()
    config.option.ptmproc_config.num_cores = mproc_num_cores
    config.option.ptmproc_config.max_simultaneous_connections = mproc_max_connections or 10
    # this is what pycov uses to determine we are distributed:
    config.option.numprocesses = config.option.ptmproc_config.num_cores
    if config.option.numprocesses and config.option.numprocesses < 0:
        raise pytest.UsageError("Number of cores must be greater than or equal to zero when running as a master")
    if config.option.ptmproc_config.max_simultaneous_connections and \
            config.option.ptmproc_config.max_simultaneous_connections <= 0:
        raise pytest.UsageError("max simultaneous connections must be greater than 0; preferably greater than 9")
    # validation
    if config.option.ptmproc_config.num_cores is None or is_degraded() or mproc_disabled:
        if '--as-main' in sys.argv or '--as-worker' in sys.argv:
            raise pytest.UsageError(
                "Refusing to execute on distributed system without also specifying the number of cores to use")
        reporter.write(
            ">>>>> no number of cores provided or running in environment unsupportive of parallelized testing, "
            "not running multiprocessing <<<<<\n", yellow=True)
        return
    if getattr(config.option, "mproc_client_connect", None):
        mproc_pytest_cmdline_coordinator(config)
    else:
        mproc_pytest_cmdline_main(config, reporter=reporter)


def mproc_pytest_cmdline_coordinator(config):
    from pytest_mproc.orchestration import OrchestrationManager
    from pytest_mproc.coordinator import CoordinatorFactory
    config.option.no_summary = True
    config.option.no_header = True
    config.option.ptmproc_config = PytestMprocConfig(
        num_cores=config.option.ptmproc_config.num_cores,
        server_uri=None,
        client_connect=getattr(config.option, "mproc_client_connect", None)
    )
    if config.option.ptmproc_config.num_cores < 1:
        raise pytest.UsageError("Number of cores must be 1 or more when running as client")
    if config.option.ptmproc_config.client_connect and not config.option.worker:
        uri = config.option.ptmproc_config.client_connect
        debug_print(f"Running coordinator connecting to {uri}")
        config.option.ptmproc_config.server_uri = uri
        if not config.option.collectonly:
            mgr = OrchestrationManager.create(uri=uri, as_client=True)
            factory = CoordinatorFactory(
                num_processes=config.option.ptmproc_config.num_cores,
                mgr=mgr,
                as_remote_client=True)
            config.ptmproc_runtime = PytestMprocRuntime(mproc_main=None,
                                                        coordinator=factory.launch(uri))


def mproc_pytest_cmdline_main(config, reporter: BasicReporter):
    from pytest_mproc.main import Orchestrator
    from pytest_mproc.coordinator import CoordinatorFactory
    assert "--as-worker" not in sys.argv
    is_server = hasattr(config.option, 'mproc_server_uri') and config.option.mproc_server_uri is not None
    config.option.ptmproc_config.server_uri = getattr(config.option, 'mproc_server_uri') or \
                                              f'127.0.0.1:{find_free_port()}'
    local_proj_file = Path("./ptmproc_project.cfg")
    project_config = getattr(config.option, "project_structure_path", None) or \
        (local_proj_file if local_proj_file.exists() else None)
    if is_server:
        remote_clients = config.option.mproc_remote_clients
        version = str(sys.version_info[0]) + "." + str(sys.version_info[1])
        mproc_remote_sys_executable = getattr(config.option, "mproc_remote_sys_executable", None) or \
            f"/usr/bin/python{version}"
        if bool(project_config) != bool(remote_clients):
            # basically if only one is None
            raise pytest.UsageError(f"Must specify both project config and remote clients together or not at all "
                                    f"{project_config} {remote_clients}")
        config.option.ptmproc_config.remote_hosts = None if not remote_clients \
            else RemoteHostConfig.from_list(remote_clients)\
            if isinstance(remote_clients, Iterable)\
            else RemoteHostConfig.from_uri_string(remote_clients)
        config.option.ptmproc_config.remote_sys_executable = mproc_remote_sys_executable
    if config.option.ptmproc_config.num_cores < 1:
        raise pytest.UsageError("Number of cores must be 1 or more when running on single host")
    reporter.write(f"Running as main @ {config.option.ptmproc_config.server_uri}\n", green=True)
    default_proj_config = Path(os.getcwd()) / "ptmproc_project.cfg"
    if is_server:
        project_config = project_config if (project_config is not None or not default_proj_config.exists())\
            else default_proj_config
    else:
        project_config = None
    orchestrator = Orchestrator(
        deploy_timeout=DEPLOY_TIMEOUT,
        remote_sys_executable=config.option.ptmproc_config.remote_sys_executable,
        project_config=ProjectConfig.from_file(Path(project_config)) if project_config else None,
        uri=config.option.ptmproc_config.server_uri,
        remote_clients_config=config.option.ptmproc_config.remote_hosts)
    config.ptmproc_runtime = PytestMprocRuntime(mproc_main=orchestrator, coordinator=None)
    if not is_server:
        # noinspection PyProtectedMember
        factory = CoordinatorFactory(
            num_processes=config.option.ptmproc_config.num_cores,
            mgr=config.ptmproc_runtime.mproc_main._mp_manager,
            as_remote_client=False)
        coordinator = factory.launch(config.option.ptmproc_config.server_uri)
        config.ptmproc_runtime.coordinator = coordinator
    config.option.dist = "no"
    val = config.getvalue
    if not val("collectonly"):
        usepdb = config.getoption("usepdb")  # a core option
        if val("dist") != "no" and usepdb:
            raise pytest.UsageError(
                "--pdb is incompatible with distributing tests."
            )  # noqa: E501


def pytest_sessionstart(session):
    with suppress(Exception):
        reporter = session.config.pluginmanager.getplugin('terminalreporter')
        if reporter:
            reporter._session = session


@pytest.mark.trylast
def pytest_configure(config):
    if not config.option.worker and \
            (config.option.ptmproc_config.num_cores is None or getattr(config.option, "mproc_disabled")):
        return  # return of None indicates other hook impls will be executed to do the task at hand
    # tell xdist not to run, (and BTW setting numprocesses is enough to tell pycov we are distributed)
    config.option.dist = "no"


def process_fixturedef(item, target, fixturedef, request, config, scope):
    if not fixturedef:
        return
    for argname in fixturedef.argnames:
        if argname in config.option.fixtures:
            continue
        # noinspection PyProtectedMember
        fixdef = item._fixtureinfo.name2fixturedefs.get(argname, [None])[0]
        if fixdef is None:
            continue
        if fixturedef.scope in ['node', 'global']:
            process_fixturedef(item, target, fixdef, request, config, fixdef.scope)
    generated = config.generated_fixtures
    if not fixturedef or fixturedef.scope != scope:
        return
    name = fixturedef.argname
    if name not in config.option.fixtures:
        fixture = pytest.fixtures.resolve_fixture_function(fixturedef, request)
        args = [config.option.fixtures[argname] for argname in fixturedef.argnames]
        val = fixture(*args)
        if inspect.isgenerator(val):
            generated.append(val)
            val = next(val)
        target.put_fixture(name, val)
        config.option.fixtures[name] = val


# noinspection PyProtectedMember
def pytest_runtestloop(session):
    if session.config.option.collectonly:
        return
    reporter: Optional[BasicReporter, _pytest.terminal.TerminalReporter] = \
        session.config.pluginmanager.getplugin('terminalreporter')
    if reporter:
        reporter.tests_count = len(session.items)
    if reporter is None:
        reporter = BasicReporter()
    if session.testsfailed and not session.config.option.continue_on_collection_errors:
        raise session.Interrupted("%d errors during collection" % session.testsfailed)

    if session.config.option.collectonly:
        return  # should never really get here, but for consistency
    if len(session.items) == 0:
        return
    if not session.config.option.worker and session.config.option.ptmproc_config.num_cores is None \
            or is_degraded() or getattr(session.config.option, "mproc_disabled"):
        # return of None indicates other hook impls will be executed to do the task at hand
        # aka, let the basic hook handle it from here, no distributed processing to be done
        return
    if session.shouldfail:
        if session.config.ptmproc_runtime.coordinator is not None:
            session.config.ptmproc_runtime.coordinator.kill()
        if session.config.ptmproc_runtime.mproc_main is not None:
            session.config.ptmproc_runtime.mproc_main.kill()
        raise session.Failed(session.shouldfail)

    session.config.generated_fixtures = []

    if not hasattr(session.config.option, "fixtures"):
        session.config.option.fixtures = {}
    if not session.config.option.worker \
            and session.config.ptmproc_runtime \
            and session.config.ptmproc_runtime.coordinator:
        coordinator = session.config.ptmproc_runtime.coordinator
        for item in session.items:
            if session.shouldfail:
                break
            if hasattr(item, "_request"):
                for name in item._fixtureinfo.argnames:
                    try:
                        fixturedef = item._fixtureinfo.name2fixturedefs.get(name, [None])[0]
                        if not fixturedef or fixturedef.scope != 'node':
                            continue
                        if hasattr(fixturedef.func, "_pytest_group"):
                            if hasattr(item._pyfuncitem.obj, "_pytest_group"):
                                group2 = item._pyfuncitem.obj._pytest_group
                                if group2 == fixturedef.func._pytest_group:
                                    BasicReporter().write(
                                        f"WARNING: '{item.nodeid}' specifies a group but also belongs " +
                                        f"to fixture '{fixturedef.argname}'' which specifies the same group")
                                else:
                                    raise raise_usage_error(
                                        session,
                                        f"test {item.nodeid} belongs to group '{group2.name}' but also belongs " +
                                        f"to fixture '{fixturedef.argname}' which specifies " +
                                        f"group '{fixturedef.func._pytest_group.name}'.  A test cannot" +
                                        "belong to two distinct groups")
                            item._pyfuncitem._pytest_group = fixturedef.func._pytest_group
                        if coordinator and name not in session.config.option.fixtures:
                            process_fixturedef(item, coordinator, fixturedef, item._request, session.config, 'node')
                    except Exception as e:
                        session.shouldfail = \
                            f"Exception in fixture: {e.__class__.__name__} raised with msg '{str(e)}' {format_exc()}"
                        reporter.write(f">>> Fixture ERROR: {format_exc()}\n", red=True)
                        break
    if not session.config.option.worker \
            and session.config.ptmproc_runtime\
            and session.config.ptmproc_runtime.mproc_main:
        orchestrator = session.config.ptmproc_runtime.mproc_main
        try:
            with orchestrator:
                orchestrator.set_items(session.items)
                orchestrator.run_loop(session)
        except Exception as e:
            reporter.write(format_exc() + "\n", red=True)
            reporter.write(f"\n>>> ERROR in run loop;  unexpected Exception\n {str(e)}\n\n", red=True)
            return False
    if session.config.option.worker:
        session.config.option.worker.test_loop(session)
    elif session.config.ptmproc_runtime.coordinator:
        session.config.ptmproc_runtime.coordinator.join()
    return True


@pytest.hookimpl(hookwrapper=True)
def pytest_collection_finish(session) -> None:
    config = session.config
    verbose = config.option.verbose
    if config.option.worker \
            or (config.ptmproc_runtime and not config.ptmproc_runtime.mproc_main):
        config.option.verbose = -2
    if session.config.option.worker:
        session.config.option.worker.pytest_collection_finish(session)
    yield
    config.option.verbose = verbose


def pytest_internalerror(excrepr):
    from pytest_mproc.worker import WorkerSession
    if WorkerSession.singleton():
        return WorkerSession.singleton().pytest_internalerror(excrepr)


@pytest.mark.tryfirst
def pytest_runtest_logreport(report):
    from pytest_mproc.worker import WorkerSession
    if WorkerSession.singleton():
        return WorkerSession.singleton().pytest_runtest_logreport(report)


# noinspection PyUnusedLocal
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_terminal_summary(terminalreporter, exitstatus, config):
    verbose = config.option.verbose
    if config.option.worker:
        config.option.tbstyle = 'no'
        terminalreporter.reportchars = ""
    elif config.ptmproc_runtime and not config.ptmproc_runtime.mproc_main:
        config.option.verbose = -2
    elif config.ptmproc_runtime and not config.ptmproc_runtime.mproc_main:
        config.option.tbstyle = 'no'
        terminalreporter.reportchars = ""
        if user_output.verbose:
            terminalreporter.line(">>> This is a satellite processing node. "
                                  "Please see master node output for actual test summary <<<<<<<<<<<<",
                                  yellow=True)
    yield
    if config.option.worker:
        config.option.verbose = verbose


# noinspection PyUnusedLocal
@pytest.hookimpl(hookwrapper=True)
def pytest_report_header(config, startdir):
    verbose = config.option.verbose
    if config.option.worker \
            or (config.ptmproc_runtime and not config.ptmproc_runtime.mproc_main):
        config.option.verbose = -2
    yield
    config.option.verbose = verbose


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_sessionstart(session) -> None:
    verbose = session.config.option.verbose
    if session.config.option.worker\
            or (session.config.ptmproc_runtime and not session.config.ptmproc_runtime.mproc_main):
        session.config.option.verbose = -2
    yield
    session.config.option.verbose = verbose


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_sessionfinish(session):
    verbose = session.config.option.verbose
    with suppress(Exception):
        fixtures.Node.Manager.shutdown()
    with suppress(Exception):
        fixtures.Global.Manager.shutdown()
    if session.config.option.worker:
        session.config.option.verbose = -2
    if session.config.ptmproc_runtime and not session.config.ptmproc_runtime.mproc_main:
        session.config.option.verbose = -2
    if not session.config.option.worker and session.config.ptmproc_runtime\
            and session.config.ptmproc_runtime.coordinator is not None:
        with suppress(Exception):
            session.config.ptmproc_runtime.coordinator.kill()
    if not session.config.option.worker and session.config.ptmproc_runtime\
            and session.config.ptmproc_runtime.mproc_main is not None:
        orchestrator = session.config.ptmproc_runtime.mproc_main
        orchestrator.shutdown()
    yield
    session.config.option.verbose = verbose


def raise_usage_error(session, msg: str):
    if session.config.ptmproc_runtime.coordinator is not None:
        session.config.ptmproc_runtime.coordinator.kill()
    raise pytest.UsageError(msg)

################
# Process-safe temp dir
###############


class TmpDirFactory:

    """
    tmpdir is not process/thread safe when used in a multiprocessing environment.  Failures on setup can
    occur (even if infrequently) under certain rae conditoins.  This provides a safe mechanism for
    creating temporary directories utilizng s a global-scope fixture
    """

    def __init__(self):
        self._root_tmp_dir = tempfile.mkdtemp(prefix=f"pytest_mproc-{getpass.getuser()}-{os.getpid()}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with suppress(Exception):
            shutil.rmtree(self._root_tmp_dir)

    @contextmanager
    def create_tmp_dir(self, cleanup_immediately: bool = True):
        """
        :param cleanup_immediately: if True, rm the directory and all contents when associated fixture is no longer
           in use, otherwise wait until end of test session when everything is cleaned up
        :return: newly create temp directory unique across all Process's
        """
        tmpdir = tempfile.mkdtemp(dir=self._root_tmp_dir)
        try:
            yield Path(tmpdir)
        finally:
            if cleanup_immediately:
                with suppress(Exception):
                    shutil.rmtree(tmpdir)


@pytest.fixture(scope='session')
def mp_tmp_dir_factory():
    """
    :return: a factory for creating unique tmp directories, unique across all Process's
    """
    with TmpDirFactory() as factory:
        yield factory


@pytest.fixture()
def mp_tmp_dir(mp_tmp_dir_factory: TmpDirFactory):
    # tmpdir is not thread safe and can fail on test setup when running on a highly loaded very parallelized system
    # so use this instead
    with mp_tmp_dir_factory.create_tmp_dir("PYTEST_MPROC_LAZY_CLEANUP" not in os.environ) as tmp_dir:
        yield tmp_dir



"""
This package contains code to coordinate execution from a main thread to worker threads (processes)
"""
import os
import sys
import time
from contextlib import suppress

from multiprocessing.managers import SyncManager
from subprocess import TimeoutExpired
from typing import Optional
from pytest_mproc import find_free_port, get_ip_addr, get_auth_key, FatalError
from pytest_mproc.utils import BasicReporter


class CoordinatorFactory:
    """
    Responsible for creating Coordinator instances and launching them
    """

    sm: Optional[SyncManager] = None

    def __init__(self, num_processes: int, mgr, as_remote_client: bool):
        """
        :param num_processes: number of parallel executions to be conducted
        :param mgr: global manager of queues, etc.
        :param as_remote_client: whether running as remote client (or on local host if False)
        """
        self._num_processes = num_processes
        self._is_local = not as_remote_client
        if self.sm is None and not self._is_local:
            self.sm = SyncManager(authkey=get_auth_key(), address=(get_ip_addr(), find_free_port()))
            self.sm.start()
        self._mgr = mgr

    def launch(self, uri: str) -> "Coordinator":
        if not self._is_local:
            # noinspection PyUnresolvedReferences
            coordinator = self.sm.Coordinator(self._num_processes,
                                              self._is_local)
        else:
            coordinator = Coordinator(self._num_processes,
                                      self._is_local)
        # noinspection PyUnresolvedReferences
        self._mgr.register_client(coordinator)
        executable = os.environ.get('PTMPROC_EXECUTABLE', sys.executable)
        coordinator.start(uri, executable)
        return coordinator


class Coordinator:
    """
    Context manager for kicking off worker Processes to conduct test execution via pytest hooks
    Coordinators are scoped to a node and only handle a collection of workers on that node
    """

    def __init__(self, num_processes: int, is_local: bool):
        """
        :param num_processes: number of parallel executions to be conducted
        :param is_local: whether Coordinator is to be run on same node as main orchestrator or is on a remote machine
        """
        self._is_local = is_local
        self._num_processes = num_processes
        self._count = 0
        self._session_start_time = time.time()
        self._reporter = BasicReporter()
        self._worker_procs = []

    def is_local(self):
        return self._is_local

    def start(self, uri: str, executable: str) -> None:
        """
        Start all worker processes

        :return: this object
        """
        from pytest_mproc.worker import WorkerSession
        for index in range(self._num_processes):
            proc = WorkerSession.start(uri, executable)
            self._worker_procs.append(proc)
        time.sleep(3)
        failed = 0
        for proc in self._worker_procs:
            if proc.returncode is not None:
                failed += 1
        if failed == len(self._worker_procs):
            raise FatalError("all worker client have died unexpectedly")
        return

    def join(self, timeout: Optional[float] = None):
        for proc in self._worker_procs:
            try:
                proc.wait(timeout)
            except TimeoutExpired:
                with suppress(OSError):
                    proc.kill()

    def kill(self):
        if CoordinatorFactory.sm:
            CoordinatorFactory.sm.shutdown()
        for proc in self._worker_procs:
            proc.kill()
        self._worker_procs = []


# register the proxy class as the Coordinator class for SyncManager
SyncManager.register("Coordinator", Coordinator)



import os

import aiohttp
import json
import pytest
import pytest_mproc

from aiofile import async_open
from pathlib import Path
from typing import Optional, Dict, List, Iterable, Tuple, AsyncGenerator, Any
from dataclasses import dataclass, field

SRC_PATHS = 'src_paths'
REQUIREMENTS_PATHS = 'requirements_paths'
TEST_FILES = 'test_files'


@dataclass
class ProjectConfig:
    project_name: str
    project_root: Path
    test_files: List[Path]
    artifcats_path: Optional[Path] = Path('./artifacts')
    src_paths: List[Path] = field(default_factory=list)
    prepare_script: Optional[Path] = None
    finalize_script: Optional[Path] = None

    @classmethod
    def from_file(cls, path: Path) -> "ProjectConfig":
        with open(path, 'rb') as in_stream:
            try:
                data = json.load(in_stream)
                if 'artifacts_path' in data:
                    if Path(data['artifacts_path']).is_absolute():
                        raise pytest.UsageError(
                            f"'artifacts_dir' in project config '{str(path)}' must be a relative path")
                if 'project_name' not in data:
                    raise pytest.UsageError(f"project config {str(path)} does not define 'project_name'")
                if not data['project_name'].strip():
                    raise pytest.UsageError(f"project_name in {str(path)} cannot be empty string")
                if 'finalize_script' in data and (not Path(data['finalize_script']).exists or
                                                  not os.access(data['finalize_script'], os.X_OK)):
                    raise pytest.UsageError(
                        f"finalize script {data['finalize_script']} either does not exist or is not executable")
                if 'prepare_script' in data and (not Path(data['prepare_script']).exists or
                                                 not os.access(data['prepare_script'], os.X_OK)):
                    raise pytest.UsageError(
                        f"prepare script {data['prepare_script']} either does not exist or is not executable")
                if TEST_FILES not in data:
                    raise pytest.UsageError(f"'{TEST_FILES}' is not in project config '{path}'")
                if type(data[TEST_FILES]) != list:
                    raise pytest.UsageError(f"{TEST_FILES} in project config '{path}' is not a list of paths")
                if any([type(f) != str for f in data[TEST_FILES]]):
                    raise pytest.UsageError(f"'test_path' must be a single string path in '{path}'")
                if SRC_PATHS in data and not isinstance(data[SRC_PATHS], Iterable):
                    raise pytest.UsageError(f"'{SRC_PATHS}' in project config '{path}' is not a list of string paths")
                for p in data[SRC_PATHS]:
                    if not (path.parent / p).exists():
                        raise pytest.UsageError(f"source path '{p}' relative to project config "
                                                f"{str(path)} is not a path")
                return ProjectConfig(
                    project_name=data['project_name'],
                    project_root=path.parent,
                    test_files=[Path(p) for p in data[TEST_FILES]],
                    src_paths=[Path(p) for p in data[SRC_PATHS]],
                    prepare_script=data.get('prepare_script'),
                    artifcats_path=data.get('artifacts_path', Path("./artifacts")),
                    finalize_script=data.get('finalize_script'),
                )
            except json.JSONDecodeError:
                raise pytest.UsageError(f"Invalid json format in project config '{path}'")
            except (KeyError, ValueError)as e:
                raise pytest.UsageError(f"Invalid json data in project config '{path}': {e}")


@dataclass
class RemoteHostConfig:
    remote_host: str
    arguments: Dict[str, str] = field(default_factory=dict)
    remote_root: Optional[Path] = None

    @classmethod
    def from_list(cls, value: Iterable[str]) -> List["RemoteHostConfig"]:
        """
        parse a cli input string for remote host and parameters specifications
        :param value: a string spec, a str path to a file in form file://..., or to a url in form 'http[s]://...'
            (the latter two returning a json dict in correspondence with return)
        :return: dictionary keyed on host of a dictionary of key value parameters
        """
        configs: List["RemoteHostConfig"] = []
        try:
            for item in value:
                host, *args = item.split(';')
                configs.append(RemoteHostConfig(
                    remote_host=host,
                    arguments={k: v for k, v in [arg.split('=') for arg in args]}
                ))
        except ValueError:
            raise pytest.UsageError(f"invalid cli input for remote hosts: {value}")
        return configs

    @classmethod
    async def from_uri_string(cls, value: str) -> AsyncGenerator["RemoteHostConfig", "RemoteHostConfig"]:
        """
        parse a cli input string for remote host and parameters specifications
        :param value: a string spec, a str path to a file in form file://..., or to a url in form 'http[s]://...'
            (the latter two returning a json dict in correspondence with return)
        :return: dictionary keyed on host of a dictionary of key value parameters
        """
        def validate(json_data: Dict[str, Any]):
            if type(json_data) != dict:
                raise pytest.UsageError(
                    f"Invalid json format for client host/args specification from '{value}'"
                )
            if 'host' not in json_data:
                raise pytest.UsageError("'host' not found in provided json")
            if 'arguments' not in json_data:
                raise pytest.UsageError("'arguments' not found in provided json")
            if type(json_data['host']) != str:
                raise pytest.UsageError(
                    f"Invalid json format;  host is not a string name"
                )
            if type(json_data['arguments']) != dict:
                raise pytest.UsageError(
                    f"Invalid json format for client host/args specification from '{value}'")
            for k, v in json_data['arguments'].items():
                if type(k) != str:
                    raise pytest.UsageError(
                        f"Invalid json format for client host/args specification '{value}'")
                if type(v) not in (str, bool, int, float):
                    raise pytest.UsageError(
                        f"Invalid json format for client host/args specification '{value}'")
        try:
            if value.startswith('http://') or value.startswith('https://'):
                async with aiohttp.ClientSession() as session:
                    async with session.get(value) as resp:
                        async for data, _ in resp.content.iter_chunks():
                            buffer += data
                            if '\n' in data:
                                *lines, buffer = data.split('\n')
                                for line in lines:
                                    if not line.strip():
                                        continue
                                    json_data = json.loads(line.strip())
                                    validate(json_data)
                                    for k, v in json_data['arguments'].items():
                                        json_data['arguments'][k] = str(v)
                                    yield RemoteHostConfig(remote_host=json_data['host'],
                                                           arguments=json_data['arguments'])
            elif value.startswith('file://'):
                async with async_open(value[7:], "r") as afp:
                    async for line in afp:
                        if line.startswith('#') or not line.strip():
                            continue
                        json_data = json.loads(line.strip())
                        validate(json_data)
                        for k, v in json_data['arguments'].items():
                            json_data['arguments'][k] = str(v)
                        yield RemoteHostConfig(remote_host=json_data['host'],
                                               arguments=json_data['arguments'])
            else:
                raise pytest.UsageError(f"Invalid value for remote host(s): {value}")
        except json.JSONDecodeError:
            raise pytest.UsageError(f"invalid json input from source '{value}'")
        except ValueError:
            raise pytest.UsageError(f"invalid cli input for remote hosts: {value}")

    def argument_env_list(self) -> Tuple[List[str], Dict[str, str]]:
        arg_list: List[str] = []
        env_list: Dict[str, str] = {}
        for name, value in self.arguments.items():
            if name.upper() == name:
                env_list[name] = f"\"{str(value)}\""
            else:
                arg_list.append(f"--{name}")
                arg_list.append(f"\"{str(value)}\"")
        return arg_list, env_list


@dataclass
class PytestMprocConfig:
    num_cores: int = 1
    server_uri: Optional[str] = 'localhot:None'
    client_connect: Optional[str] = None
    connection_timeout: int = 30
    remote_hosts: List[RemoteHostConfig] = field(default_factory=list)
    max_simultaneous_connections = 24
    remote_sys_executable: Optional[Path] = None


@dataclass
class PytestMprocWorkerRuntime:
    @classmethod
    def is_worker(cls):
        return True

    coordinator: Optional["pytest_mproc.coordinator.Coordinator"] = None


@dataclass
class PytestMprocRuntime:

    def is_worker(self):
        return self.mproc_main is None

    mproc_main: Optional["pytest_mproc.main.Orchestrator"] = None
    coordinator: Optional["pytest_mproc.coordinator.Coordinator"] = None



import os
import socket
import subprocess
import sys
import time
import traceback
from contextlib import suppress
from multiprocessing import (
    JoinableQueue,
    Queue,
    Semaphore,
)
from typing import Iterator, Union
import pytest
# noinspection PyProtectedMember
from _pytest.config import _prepareconfig
import resource

from pytest_mproc import resource_utilization, TestError, user_output, get_ip_addr, get_auth_key
from pytest_mproc.data import (
    ClientDied,
    ResultExit,
    ResourceUtilization,
    TestBatch,
    TestState,
    TestStateEnum,
)
from pytest_mproc.data import ResultException, ResultTestStatus, ResultType
from pytest_mproc.ptmproc_data import (
    PytestMprocConfig,
    PytestMprocWorkerRuntime,
)
from pytest_mproc.user_output import debug_print
from pytest_mproc.utils import BasicReporter

if sys.version_info[0] < 3:
    # noinspection PyUnresolvedReferences
    from Queue import Empty
else:
    # noinspection PyUnresolvedReferences
    from queue import Empty

"""maximum time between reporting status back to coordinator"""
MAX_REPORTING_INTERVAL = 1.0  # seconds


class WorkerSession:
    """
    Handles reporting of test status and the like
    """

    def __init__(self, index, is_remote: bool, test_q: JoinableQueue, result_q: Queue):
        self._is_remote = is_remote
        self._this_host = get_ip_addr()
        self._index = index
        self._name = "worker-%d" % (index + 1)
        self._count = 0
        self._session_start_time = time.time()
        self._buffered_results = []
        self._buffer_size = 20
        self._timestamp = time.time()
        self._last_execution_time = time.time()
        self._resource_utilization = ResourceUtilization(-1.0, -1.0, -1.0, -1)
        self._reporter = BasicReporter()
        self._test_q = test_q
        self._result_q = result_q
        self._ip_addr = get_ip_addr()
        self._pid = os.getpid()

    def _put(self, result: Union[TestStateEnum, ResultType, ResultExit], timeout=None):
        """
        Append test result data to queue, flushing buffered results to queue at watermark level for efficiency

        :param result: test result data
        :param timeout: timeout after this many seconds, if specified
        :raises: TimeoutError is not executed in time
        """
        if self._is_remote and isinstance(result, ResultTestStatus):
            os.write(sys.stderr.fileno(), b'.')
        if isinstance(result, ResultExit):
            self._result_q.put(result)
        else:
            self._buffered_results.append(result)
        if isinstance(result, (TestStateEnum, ResultExit)) \
                or len(self._buffered_results) >= self._buffer_size \
                or (time.time() - self._timestamp) > MAX_REPORTING_INTERVAL:
            self._flush(timeout)

    def _flush(self, timeout=None):
        """
        fluh buffered results out to the queue.
        """
        if self._buffered_results:
            if timeout is not None:
                self._result_q.put(self._buffered_results, timeout)
            else:
                self._result_q.put(self._buffered_results)
            self._buffered_results = []
            self._timestamp = time.time()

    def test_loop(self, session):
        """
        This is where the action takes place.  We override the usual implementation since
        that doesn't support a dynamic generation of tests (whose source is the test Queue
        that it draws from to pick out the next test)

        :param session:  Where the tests generator is kept
        """
        # global my_cov
        start_time = time.time()
        rusage = resource.getrusage(resource.RUSAGE_SELF)
        try:
            if session.testsfailed and not session.config.option.continue_on_collection_errors:
                raise session.Failed("%d errors during collection" % session.testsfailed)

            if session.config.option.collectonly:
                return  # should never really get here, but for consistency
            for test_batch in session.items_generator:
                if test_batch is None:
                    break
                # test item comes through as a unique string nodeid of the test
                # We use the pytest-collected mapping we squirrelled away to look up the
                # actual _pytest.python.Function test callable
                # NOTE: session.items is an iterator, as wet upon construction of WorkerSession
                for test_id in test_batch.test_ids:
                    has_error = False
                    bare_test_id = test_id.split(os.sep)[-1]
                    item = session._named_items[bare_test_id]
                    self._put(TestState(TestStateEnum.STARTED, self._this_host, self._pid, test_id, test_batch))
                    try:
                        item.config.hook.pytest_runtest_protocol(item=item, nextitem=None)
                    except TestError:
                        has_error = True
                        self._put(TestState(TestStateEnum.RETRY, self._this_host, self._pid, test_id, test_batch))
                    except Exception as e:
                        import traceback
                        raise Exception(f"Exception running test: {e}: {traceback.format_exc()}")

                    finally:
                        if not has_error:
                            self._put(TestState(TestStateEnum.FINISHED, self._this_host, self._pid,
                                                test_id, test_batch))
                    # very much like that in _pytest.main:
                    try:
                        if session.shouldfail:
                            raise session.Failed(session.shouldfail)
                    except AttributeError:
                        pass  # some version of pytest do not have this attribute
                    if session.shouldstop:
                        raise session.Interrupted(session.shouldstop)
                    # count tests that have been run
                    self._count += 1
                    self._last_execution_time = time.time()
        finally:
            self._flush()
            assert self._buffered_results == []
            end_usage = resource.getrusage(resource.RUSAGE_SELF)
            time_span = self._last_execution_time - start_time
            self._resource_utilization = resource_utilization(time_span=time_span,
                                                              start_rusage=rusage,
                                                              end_rusage=end_usage)
            with suppress(Exception):
                self.session_finish()

    @pytest.mark.tryfirst
    def pytest_collection_finish(self, session):
        """
        Invoked once pytest has collected all information about which tests to run.
        Those items are squirrelled away in a different attributes, and a generator
        is put in its place to draw from.

        :param session: the pytest session

        :return: the generator of tests to run
        """
        assert self._test_q is not None

        def generator(test_q: JoinableQueue) -> Iterator[TestBatch]:
            test = test_q.get()
            while test:
                test_q.task_done()
                yield test
                test = test_q.get()
            test_q.task_done()
        session.items_generator = generator(self._test_q)
        session._named_items = {item.nodeid.split(os.sep)[-1]: item for item in session.items}
        return session.items

    @classmethod
    def start(cls, uri: str, executable: str) -> subprocess.Popen:
        """
        :param uri: uri of main server
        :param executable: which executable to run under
        :return: Process created for new worker
        """
        from pytest_mproc.fixtures import Node
        import binascii
        env = os.environ.copy()
        env["PYTEST_WORKER"] = "1"
        env["AUTH_TOKEN_STDIN"] = "1"
        if user_output.verbose:
            env['PTMPROC_VERBOSE'] = '1'
        else:
            env['PTMPROC_VERBOSE'] = '0'
        env['PTMPROC_NODE_MGR_PORT'] = str(Node.Manager.PORT)
        stdout = sys.stdout
        stderr = sys.stderr
        executable = executable or sys.executable
        proc = subprocess.Popen([executable, '-m', __name__, uri] + sys.argv[1:],
                                env=env,
                                stdout=stdout, stderr=stderr, stdin=subprocess.PIPE)
        proc.stdin.write(binascii.b2a_hex(get_auth_key()) + b'\n')
        proc.stdin.flush()
        return proc

    @staticmethod
    def run(index: int, is_remote: bool, start_sem: Semaphore, fixture_sem: Semaphore,
            test_q: JoinableQueue, result_q: Queue, sync_sem: Semaphore()) -> None:
        from pytest_mproc.orchestration import OrchestrationManager
        sync_sem.release()
        start_sem.acquire()
        worker = WorkerSession(index, is_remote, test_q, result_q)
        worker._fixture_sem = fixture_sem
        args = sys.argv[1:]
        # remove coverage args, as pytest_cov handles multiprocessing already and will apply coverage to worker
        # as a proc that was launched from main thread which itself has coverage (otherwise it will attempt
        # duplicate coverage processing and file conflicts galore)
        args = [arg for arg in args if not arg.startswith("--cov=")]
        config = _prepareconfig(args, plugins=[])
        config.option.ptmproc_config = PytestMprocConfig()
        config.ptmproc_runtime = PytestMprocWorkerRuntime()
        # unregister terminal (don't want to output to stdout from worker)
        # as well as xdist (don't want to invoke any plugin hooks from another distribute testing plugin if present)
        config.pluginmanager.unregister(name="terminal")
        config.pluginmanager.register(worker, "mproc_worker")
        worker._client = OrchestrationManager.create(uri=config.option.ptmproc_config.server_uri, as_client=True)
        workerinput = {'slaveid': "worker-%d" % worker._index,
                       'workerid': "worker-%d" % worker._index,
                       'cov_master_host': socket.gethostname(),
                       'cov_slave_output': os.path.join(os.getcwd(), "worker-%d" % worker._index),
                       'cov_master_topdir': os.getcwd()
                       }
        config.slaveinput = workerinput
        config.slaveoutput = workerinput
        try:
            # and away we go....
            config.hook.pytest_cmdline_main(config=config)
        finally:
            # noinspection PyProtectedMember
            config._ensure_unconfigure()
            worker._reporter.write(f"\nWorker-{index} finished\n")

    def pytest_internalerror(self, __excrepr):
        self._put(ResultException(SystemError("Internal pytest error")))

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_logreport(self, report):
        """
        report only status of calls (not setup or teardown), unless there was an error in those
        stash report for later end-game parsing of failures

        :param report: report to draw info from
        """
        self._put(ResultTestStatus(report))

    # noinspection PyUnusedLocal
    def session_finish(self):
        """
        output failure information and final exit status back to coordinator
        """
        self._put(ResultExit(self._index,
                             self._count,
                             0,
                             self._last_execution_time - self._session_start_time,
                             self._resource_utilization))
        self._flush()
        self.set_singleton(None)
        return True

    @classmethod
    def set_singleton(cls, singleton):
        cls._singleton = singleton

    @classmethod
    def singleton(cls):
        return cls._singleton

    _singleton = None


def pytest_cmdline_main(config):
    assert WorkerSession.singleton() is not None
    config.option.worker = WorkerSession.singleton()


def main():
    from pytest_mproc import plugin  # ensures auth_key is set
    from pytest_mproc.fixtures import Node
    # noinspection PyUnresolvedReferences
    from pytest_mproc.worker import WorkerSession  # to make Python happy
    assert plugin  # to prevent flake8 unused import
    # from pytest_mproc.worker import WorkerSession
    from pytest_mproc.orchestration import OrchestrationManager
    uri = sys.argv[1]
    mgr = OrchestrationManager.create(uri=uri, as_client=True)
    # noinspection PyUnresolvedReferences
    mgr.register_worker((get_ip_addr(), os.getpid()))
    # noinspection PyUnresolvedReferences
    test_q = mgr.get_test_queue()
    # noinspection PyUnresolvedReferences
    result_q = mgr.get_results_queue()

    worker = WorkerSession(index=os.getpid(),
                           is_remote='--as-worker' in sys.argv,
                           test_q=test_q,
                           result_q=result_q,
                           )
    WorkerSession.set_singleton(worker)
    errored = False
    msg = None
    assert WorkerSession.singleton() is not None
    status = None
    # noinspection PyBroadException
    try:
        status = pytest.main(sys.argv[2:])
        if isinstance(status, pytest.ExitCode):
            status = status.value
    except Exception as e:
        errored = True
        msg = f"Worker {get_ip_addr()}-{os.getpid()} died with exception {e}\n {traceback.format_exc()}"
        os.write(sys.stderr.fileno(), msg.encode('utf-8'))
    finally:
        with suppress(Exception):  # result q is probably closed by now, but for good measure...
            result_q.put(ClientDied(os.getpid(), get_ip_addr(), errored=errored, message=msg))
        with suppress(Exception):
            Node.Manager.shutdown()
    return status


if __name__ == "__main__":
    sys.exit(main())



from dataclasses import dataclass
from enum import Enum
from typing import List, Union, Optional

from pytest_mproc import DEFAULT_PRIORITY
from pytest import TestReport


class TestExecutionConstraint(Enum):
    __test__ = False  # tell pytest not to treat as test class
    SINGLE_PROCESS ="SINGLE_PROCESS"
    SINGLE_NODE = "SINGLE_NODE"


@dataclass
class TestBatch:
    test_ids: List[str]
    priority: int = DEFAULT_PRIORITY
    restriction: TestExecutionConstraint = TestExecutionConstraint.SINGLE_PROCESS


@dataclass
class ResourceUtilization:
    time_span: float
    user_cpu: float
    system_cpu: float
    memory_consumed: int  # Megabytes


@dataclass
class ResultException(Exception):
    excrepr: Exception


@dataclass
class ClientDied(Exception):
    pid: int
    host: str
    errored: bool = False
    message: Optional[str] = None


@dataclass
class AllClientsCompleted:
    pass


@dataclass
class ResultTestStatus:
    report: TestReport


class TestStateEnum(Enum):
    STARTED = 1
    FINISHED = 2
    RETRY = 3


@dataclass
class TestState:
    state: TestStateEnum
    host: str
    pid: int
    test_id: str
    test_batch: TestBatch

@dataclass
class ResultExit:
    worker_index: int
    test_count: int
    status: int
    duration: float
    resource_utilization: ResourceUtilization


@dataclass
class GroupTag:
    name: str
    priority: int = DEFAULT_PRIORITY
    restrict_to: TestExecutionConstraint = TestExecutionConstraint.SINGLE_PROCESS

    def __hash__(self):
        return self.name.__hash__()


ResultType = Union[TestState, ResultException, ResultExit, ResourceUtilization, ResultTestStatus]



import multiprocessing
from multiprocessing.process import current_process

import binascii
import inspect
import os
import secrets
import socket
import sys
from contextlib import closing
from typing import Union, Callable, Optional, Any, Dict

# to indicate if running as main (host) or not;  set in plugin.py
is_main = '--as-main' in sys.argv
is_worker = '--as-worker' in sys.argv

_auth_key = None
_user_defined_port_alloc: Optional[Callable[[], int]] = None
_user_defined_auth: Optional[Callable[[], bytes]] = None


def set_user_defined_auth_tokn(func: Callable[[], bytes]):
    _user_defined_auth = func


def set_user_defined_port_alloc(func: Callable[[], int]):
    _user_defined_port_alloc = func


def get_auth_key() -> bytes:
    global _auth_key
    if hasattr(multiprocessing, 'parent_process') and multiprocessing.parent_process() is not None:
        _auth_key = current_process().authkey
    elif _auth_key is not None:
        return _auth_key
    elif _user_defined_auth:
        _auth_key = _user_defined_auth()
    elif 'AUTH_TOKEN_STDIN' in os.environ:
        auth_key = sys.stdin.readline().strip()
        _auth_key = binascii.a2b_hex(auth_key)
    else:
        _auth_key = secrets.token_bytes(64)
    current_process().authkey = _auth_key
    return _auth_key


def get_auth_key_hex():
    return binascii.b2a_hex(get_auth_key()).decode('utf-8')


class TestError(Exception):
    """
    To be thrown when a test setup/test environment issue occurs preventing test execution from even happening
    """
    pass


def priority(level: int):

    def decorator_priority(object):
        if inspect.isclass(object):
            for method in [getattr(object, m) for m in dir(object) if
                           inspect.isfunction(getattr(object, m)) and m.startswith('test')]:
                if not hasattr(method, "_pytest_priority"):
                    method._pytest_priority = level
        elif inspect.isfunction(object):
            object._pytest_priority = level
        else:
            raise Exception("group decorator can only decorate class or function object")
        return object

    return decorator_priority


def group(tag: Union["GroupTag", str]):
    """
    Decorator for grouping tests or a class of tests
    :param tag: unique name for group of tests to be serialized under execution
    """
    from pytest_mproc.data import GroupTag

    def decorator_group(object):
        if inspect.isclass(object):
            for method in [getattr(object, m) for m in dir(object) if inspect.isfunction(getattr(object, m)) and m.startswith('test')]:
                method._pytest_group = tag if isinstance(tag, GroupTag) else GroupTag(tag)
        elif inspect.isfunction(object):
            object._pytest_group = tag
        else:
            raise Exception("group decorator can only decorate class or function object")
        return object
    return decorator_group


def resource_utilization(time_span: float, start_rusage, end_rusage) -> "ResourceUtilization":
    from pytest_mproc.data import ResourceUtilization
    if time_span <= 0.001:
        return -1, -1, -1, end_rusage.ru_maxrss
    if sys.platform.lower() == 'darwin':
        # OS X is in bytes
        delta_mem = (end_rusage.ru_maxrss - start_rusage.ru_maxrss) / 1000.0
    else:
        delta_mem = (end_rusage.ru_maxrss - start_rusage.ru_maxrss)
    ucpu_secs = end_rusage.ru_utime - start_rusage.ru_utime
    scpu_secs = end_rusage.ru_stime - start_rusage.ru_stime
    return ResourceUtilization(
        time_span,
        (ucpu_secs / time_span) * 100.0,
        (scpu_secs / time_span) * 100.0,
        delta_mem)


def find_free_port():
    if _user_defined_port_alloc:
        return _user_defined_port_alloc()
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def fixture(fixturefunction, *, scope, **kargs):
    from _pytest.fixtures import fixture
    return fixture(fixturefunction, scope=scope, **kargs)


def get_ip_addr():
    hostname = socket.gethostname()
    # noinspection PyBroadException
    try:
        return socket.gethostbyname(hostname)
    except Exception:
        return None



DEFAULT_PRIORITY = 10
try:
    from pytest_mproc.data import GroupTag
except:
    # for orchestrationmanager to be able to come up independently
    # TODO: fix this
    pass


class FatalError(Exception):
    """
    raised to exit pytest immediately
    """


class Constants:
    """
    For storing global "constants"
    """
    ptmproc_args: Optional[Dict[str, Any]] = None

    @classmethod
    def set(cls, name: str, typ: Any):
        cls.ptmproc_args = cls.ptmproc_args or {}
        cls.ptmproc_args[name] = typ



from pathlib import Path

import setuptools

VERSION = "5.0.12"

requirements = []
with open(Path(__file__).parent / 'pure_requirements.txt', 'r') as in_stream:
    for line in in_stream:
        line = line.strip()
        if line:
            requirements.append(line)
with open(Path(__file__).parent / 'impure_requirements.txt', 'r') as in_stream:
    for line in in_stream:
        line = line.strip()
        if line:
            requirements.append(line)

setuptools.setup(
    name='pytest_mproc',
    author='John Rusnak',
    author_email='jrusnak69@gmail.com',
    version=VERSION,
    description="low-startup-overhead, scalable, distributed-testing pytest plugin",
    package_dir={'': 'src'},
    package_data={'': ['pure_requirements.txt', 'impure_requirements.txt']},
    include_package_data=True,
    packages=setuptools.find_packages('src'),
    entry_points={
       "pytest11": ["name_of_plugin = pytest_mproc.plugin"],
    },
    classifiers=["Framework :: Pytest",
                 "Development Status :: 4 - Beta",
                 "License :: OSI Approved :: BSD License"],
    license='BSD 2-CLAUSE',
    keywords='pytest distributed multiprocessing',
    url='https://github.com/nak/pytest_mproc',
    download_url="https://github.com/nak/pytest_mproc/dist/%s" % VERSION,
    install_requires=requirements
)
